# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import xml.etree.ElementTree as ElementTree

import psrpcore.types._complex as complex
from psrpcore.types import PSChar, PSObject

from ..conftest import (
    COMPLEX_ENCODED_STRING,
    COMPLEX_STRING,
    assert_xml_diff,
    deserialize,
    serialize,
)


def test_ps_custom_object_empty():
    obj = complex.PSCustomObject()
    assert obj.PSTypeNames == ["System.Management.Automation.PSCustomObject", "System.Object"]

    obj.PSObject.to_string = "to string value"
    element = serialize(obj)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        '<TN RefId="0"><T>System.Management.Automation.PSCustomObject</T><T>System.Object</T></TN>'
        "<ToString>to string value</ToString>"
        "</Obj>"
    )

    actual = deserialize(element)
    assert isinstance(actual, complex.PSCustomObject)
    assert actual.PSTypeNames == ["System.Management.Automation.PSCustomObject", "System.Object"]
    assert str(actual) == "to string value"
    assert repr(actual) == "PSCustomObject()"


def test_ps_custom_object_type_name():
    obj = complex.PSCustomObject(**{"PSTypeName": "MyType", "My Property": "Value"})
    assert obj.PSTypeNames == ["MyType", "System.Management.Automation.PSCustomObject", "System.Object"]

    obj.PSObject.to_string = "to string value"
    element = serialize(obj)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>MyType</T>"
        "<T>System.Management.Automation.PSCustomObject</T>"
        "<T>System.Object</T></TN>"
        '<MS><S N="My Property">Value</S></MS>'
        "<ToString>to string value</ToString>"
        "</Obj>"
    )

    actual = deserialize(element)
    assert isinstance(actual, PSObject)
    assert actual.PSTypeNames == [
        "Deserialized.MyType",
        "Deserialized.System.Management.Automation.PSCustomObject",
        "Deserialized.System.Object",
    ]
    assert str(actual) == "to string value"
    assert repr(actual) == "PSObject(My Property='Value')"


def test_ps_custom_object_properties():
    obj = complex.PSCustomObject(Foo="Bar", Hello="World")
    assert obj.PSTypeNames == ["System.Management.Automation.PSCustomObject", "System.Object"]

    element = serialize(obj)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        '<TN RefId="0"><T>System.Management.Automation.PSCustomObject</T><T>System.Object</T></TN>'
        '<MS><S N="Foo">Bar</S><S N="Hello">World</S></MS>'
        "</Obj>"
    )

    actual = deserialize(element)
    assert isinstance(actual, complex.PSCustomObject)
    assert actual.PSTypeNames == ["System.Management.Automation.PSCustomObject", "System.Object"]
    assert str(actual) == "PSCustomObject(Foo='Bar', Hello='World')"
    assert repr(actual) == "PSCustomObject(Foo='Bar', Hello='World')"


def test_psrp_pipeline_result_types():
    value = complex.PipelineResultTypes.Output | complex.PipelineResultTypes.Error
    assert value.value == 3
    assert str(value) == "PipelineResultTypes.Warning"

    element = serialize(value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        "<I32>3</I32>"
        '<TN RefId="0">'
        "<T>System.Management.Automation.Runspaces.PipelineResultTypes</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>Output, Error, Warning</ToString>"
        "</Obj>"
    )

    value = deserialize(element)
    assert isinstance(value, complex.PipelineResultTypes)
    assert value == complex.PipelineResultTypes.Output | complex.PipelineResultTypes.Error


def test_console_color():
    value = complex.ConsoleColor.DarkRed
    assert value.value == 4
    assert str(value) == "ConsoleColor.DarkRed"

    element = serialize(value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        "<I32>4</I32>"
        '<TN RefId="0">'
        "<T>System.ConsoleColor</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>DarkRed</ToString>"
        "</Obj>"
    )

    value = deserialize(element)
    assert isinstance(value, complex.ConsoleColor)
    assert value == complex.ConsoleColor.DarkRed


def test_ps_thread_options():
    state = complex.PSThreadOptions.UseNewThread
    assert str(state) == "PSThreadOptions.UseNewThread"

    element = serialize(state)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        "<I32>1</I32>"
        '<TN RefId="0">'
        "<T>System.Management.Automation.Runspaces.PSThreadOptions</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>UseNewThread</ToString>"
        "</Obj>"
    )

    state = deserialize(element)
    assert isinstance(state, complex.PSThreadOptions)
    assert state == complex.PSThreadOptions.UseNewThread


def test_apartment_state():
    state = complex.ApartmentState.STA
    assert str(state) == "ApartmentState.STA"

    element = serialize(state)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        "<I32>0</I32>"
        '<TN RefId="0"><T>System.Threading.ApartmentState</T><T>System.Enum</T>'
        "<T>System.ValueType</T><T>System.Object</T></TN>"
        "<ToString>STA</ToString>"
        "</Obj>"
    )

    state = deserialize(element)
    assert isinstance(state, complex.ApartmentState)
    assert state == complex.ApartmentState.STA


def test_remote_stream_options():
    options = (
        complex.RemoteStreamOptions.AddInvocationInfoToDebugRecord
        | complex.RemoteStreamOptions.AddInvocationInfoToErrorRecord
    )
    assert str(options) == "RemoteStreamOptions.AddInvocationInfoToDebugRecord|AddInvocationInfoToErrorRecord"

    element = serialize(options)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        "<I32>5</I32>"
        '<TN RefId="0">'
        "<T>System.Management.Automation.RemoteStreamOptions</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>AddInvocationInfoToErrorRecord, AddInvocationInfoToDebugRecord</ToString>"
        "</Obj>"
    )

    options = deserialize(element)
    assert isinstance(options, complex.RemoteStreamOptions)
    assert (
        options
        == complex.RemoteStreamOptions.AddInvocationInfoToDebugRecord
        | complex.RemoteStreamOptions.AddInvocationInfoToErrorRecord
    )


def test_error_category():
    error = complex.ErrorCategory.CloseError
    assert str(error) == "ErrorCategory.CloseError"

    element = serialize(error)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        "<I32>2</I32>"
        '<TN RefId="0">'
        "<T>System.Management.Automation.ErrorCategory</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>CloseError</ToString>"
        "</Obj>"
    )

    error = deserialize(element)
    assert isinstance(error, complex.ErrorCategory)
    assert error == complex.ErrorCategory.CloseError


def test_error_record_plain():
    value = complex.ErrorRecord(
        Exception=complex.NETException("Exception"),
        CategoryInfo=complex.ErrorCategoryInfo(),
    )

    assert value.Exception.Message == "Exception"
    assert str(value) == "Exception"
    assert value.CategoryInfo.Category == complex.ErrorCategory.NotSpecified
    assert value.CategoryInfo.Activity is None
    assert value.CategoryInfo.Reason is None
    assert value.CategoryInfo.TargetName is None
    assert value.CategoryInfo.TargetType is None
    assert value.TargetObject is None
    assert value.FullyQualifiedErrorId is None
    assert value.InvocationInfo is None
    assert value.ErrorDetails is None
    assert value.PipelineIterationInfo is None
    assert value.ScriptStackTrace is None

    element = serialize(value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    expected = (
        '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Management.Automation.ErrorRecord</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<MS>"
        '<Obj RefId="1" N="Exception">'
        '<TN RefId="1">'
        "<T>System.Exception</T>"
        "<T>System.Object</T>"
        "</TN><Props>"
        '<S N="Message">Exception</S>'
        '<Nil N="Data" />'
        '<Nil N="HelpLink" />'
        '<Nil N="HResult" />'
        '<Nil N="InnerException" />'
        '<Nil N="Source" />'
        '<Nil N="StackTrace" />'
        '<Nil N="TargetSite" />'
        "</Props></Obj>"
        '<Nil N="TargetObject" />'
        '<Nil N="FullyQualifiedErrorId" />'
        '<Nil N="InvocationInfo" />'
        '<I32 N="ErrorCategory_Category">0</I32>'
        '<Nil N="ErrorCategory_Activity" />'
        '<Nil N="ErrorCategory_Reason" />'
        '<Nil N="ErrorCategory_TargetName" />'
        '<Nil N="ErrorCategory_TargetType" />'
        '<S N="ErrorCategory_Message">NotSpecified (:) [], </S>'
        '<B N="SerializeExtendedInfo">false</B>'
        "</MS>"
        "<ToString>Exception</ToString>"
        "</Obj>"
    )
    assert_xml_diff(actual, expected)

    value = deserialize(element)

    assert isinstance(value, complex.ErrorRecord)
    assert value.serialize_extended_info is False
    assert value.Exception.Message == "Exception"
    assert str(value) == "Exception"
    assert isinstance(value.CategoryInfo, complex.ErrorCategoryInfo)
    assert value.CategoryInfo.Category == complex.ErrorCategory.NotSpecified
    assert value.CategoryInfo.Activity is None
    assert value.CategoryInfo.Reason is None
    assert value.CategoryInfo.TargetName is None
    assert value.CategoryInfo.TargetType is None
    assert value.TargetObject is None
    assert value.FullyQualifiedErrorId is None
    assert value.InvocationInfo is None
    assert value.ErrorDetails is None
    assert value.PipelineIterationInfo is None
    assert value.ScriptStackTrace is None


def test_error_record_with_error_details():
    value = complex.ErrorRecord(
        Exception=complex.NETException("Exception"),
        CategoryInfo=complex.ErrorCategoryInfo(
            Category=complex.ErrorCategory.CloseError,
            Activity="Closing a file",
            Reason="File is locked",
        ),
        TargetObject="C:\\temp\\file.txt",
        FullyQualifiedErrorId="CloseError",
        ErrorDetails=complex.ErrorDetails(
            Message="Error Detail Message",
        ),
        ScriptStackTrace="At <1>MyScript.ps1",
    )

    assert value.Exception.Message == "Exception"
    assert str(value) == "Error Detail Message"
    assert value.CategoryInfo.Category == complex.ErrorCategory.CloseError
    assert value.CategoryInfo.Activity == "Closing a file"
    assert value.CategoryInfo.Reason == "File is locked"
    assert value.CategoryInfo.TargetName is None
    assert value.CategoryInfo.TargetType is None
    assert value.TargetObject == "C:\\temp\\file.txt"
    assert value.FullyQualifiedErrorId == "CloseError"
    assert value.InvocationInfo is None
    assert value.ErrorDetails.Message == "Error Detail Message"
    assert value.ErrorDetails.RecommendedAction is None
    assert value.PipelineIterationInfo is None
    assert value.ScriptStackTrace == "At <1>MyScript.ps1"

    element = serialize(value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    expected = (
        '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Management.Automation.ErrorRecord</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<MS>"
        '<Obj RefId="1" N="Exception">'
        '<TN RefId="1">'
        "<T>System.Exception</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<Props>"
        '<S N="Message">Exception</S>'
        '<Nil N="Data" />'
        '<Nil N="HelpLink" />'
        '<Nil N="HResult" />'
        '<Nil N="InnerException" />'
        '<Nil N="Source" />'
        '<Nil N="StackTrace" />'
        '<Nil N="TargetSite" />'
        "</Props>"
        "</Obj>"
        '<S N="TargetObject">C:\\temp\\file.txt</S>'
        '<S N="FullyQualifiedErrorId">CloseError</S>'
        '<Nil N="InvocationInfo" />'
        '<I32 N="ErrorCategory_Category">2</I32>'
        '<S N="ErrorCategory_Activity">Closing a file</S>'
        '<S N="ErrorCategory_Reason">File is locked</S>'
        '<Nil N="ErrorCategory_TargetName" />'
        '<Nil N="ErrorCategory_TargetType" />'
        '<S N="ErrorCategory_Message">CloseError (:) [Closing a file], File is locked</S>'
        '<S N="ErrorDetails_Message">Error Detail Message</S>'
        '<Nil N="ErrorDetails_RecommendedAction" />'
        '<S N="ErrorDetails_ScriptStackTrace">At &lt;1&gt;MyScript.ps1</S>'
        '<B N="SerializeExtendedInfo">false</B>'
        "</MS>"
        "<ToString>Error Detail Message</ToString>"
        "</Obj>"
    )
    assert_xml_diff(actual, expected)

    value = deserialize(element)

    assert isinstance(value, complex.ErrorRecord)
    assert value.serialize_extended_info is False
    assert value.Exception.Message == "Exception"
    assert str(value) == "Error Detail Message"
    assert value.CategoryInfo.Category == complex.ErrorCategory.CloseError
    assert value.CategoryInfo.Activity == "Closing a file"
    assert value.CategoryInfo.Reason == "File is locked"
    assert value.CategoryInfo.TargetName is None
    assert value.CategoryInfo.TargetType is None
    assert value.TargetObject == "C:\\temp\\file.txt"
    assert value.FullyQualifiedErrorId == "CloseError"
    assert value.InvocationInfo is None
    assert value.ErrorDetails.Message == "Error Detail Message"
    assert value.ErrorDetails.RecommendedAction is None
    assert value.PipelineIterationInfo is None
    assert value.ScriptStackTrace == "At <1>MyScript.ps1"


def test_error_record_with_invocation_info():
    value = complex.ErrorRecord(
        Exception=complex.NETException("Exception"),
        CategoryInfo=complex.ErrorCategoryInfo(),
        InvocationInfo=complex.InvocationInfo(
            BoundParameters=complex.PSDict(Path="C:\\temp\\file.txt"),
            CommandOrigin=complex.CommandOrigin.Runspace,
            ExpectingInput=False,
            HistoryId=10,
            InvocationName="Remove-Item",
            Line=10,
            OffsetInLine=20,
            PipelineLength=30,
            PipelinePosition=40,
            PositionMessage="position message",
            UnboundArguments=[True],
        ),
        PipelineIterationInfo=[1],
    )

    assert value.Exception.Message == "Exception"
    assert str(value) == "Exception"
    assert value.CategoryInfo.Category == complex.ErrorCategory.NotSpecified
    assert value.CategoryInfo.Activity is None
    assert value.CategoryInfo.Reason is None
    assert value.CategoryInfo.TargetName is None
    assert value.CategoryInfo.TargetType is None
    assert value.TargetObject is None
    assert value.FullyQualifiedErrorId is None
    assert value.InvocationInfo.BoundParameters == {"Path": "C:\\temp\\file.txt"}
    assert value.InvocationInfo.CommandOrigin == complex.CommandOrigin.Runspace
    assert value.InvocationInfo.DisplayScriptPosition is None
    assert value.InvocationInfo.ExpectingInput is False
    assert value.InvocationInfo.HistoryId == 10
    assert value.InvocationInfo.InvocationName == "Remove-Item"
    assert value.InvocationInfo.Line == "10"
    assert value.InvocationInfo.MyCommand is None
    assert value.InvocationInfo.OffsetInLine == 20
    assert value.InvocationInfo.PSCommandPath is None
    assert value.InvocationInfo.PSScriptRoot is None
    assert value.InvocationInfo.PipelineLength == 30
    assert value.InvocationInfo.PipelinePosition == 40
    assert value.InvocationInfo.PositionMessage == "position message"
    assert value.InvocationInfo.ScriptLineNumber is None
    assert value.InvocationInfo.ScriptName is None
    assert value.InvocationInfo.UnboundArguments == [True]
    assert value.ErrorDetails is None
    assert value.PipelineIterationInfo == [1]
    assert value.ScriptStackTrace is None

    element = serialize(value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    expected = (
        '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Management.Automation.ErrorRecord</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<MS>"
        '<Obj RefId="1" N="Exception">'
        '<TN RefId="1"><T>System.Exception</T>'
        "<T>System.Object</T>"
        "</TN><Props>"
        '<S N="Message">Exception</S>'
        '<Nil N="Data" />'
        '<Nil N="HelpLink" />'
        '<Nil N="HResult" />'
        '<Nil N="InnerException" />'
        '<Nil N="Source" />'
        '<Nil N="StackTrace" />'
        '<Nil N="TargetSite" />'
        "</Props>"
        "</Obj>"
        '<Nil N="TargetObject" />'
        '<Nil N="FullyQualifiedErrorId" />'
        '<Obj RefId="2" N="InvocationInfo">'
        '<TN RefId="2">'
        "<T>System.Management.Automation.InvocationInfo</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<Props>"
        '<Obj RefId="3" N="BoundParameters">'
        '<TN RefId="3">'
        "<T>System.Collections.Hashtable</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<DCT>"
        '<En><S N="Key">Path</S><S N="Value">C:\\temp\\file.txt</S></En>'
        "</DCT>"
        "</Obj>"
        '<Obj RefId="4" N="CommandOrigin">'
        "<I32>0</I32>"
        '<TN RefId="4">'
        "<T>System.Management.Automation.CommandOrigin</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>Runspace</ToString>"
        "</Obj>"
        '<Nil N="DisplayScriptPosition" />'
        '<B N="ExpectingInput">false</B>'
        '<I64 N="HistoryId">10</I64>'
        '<S N="InvocationName">Remove-Item</S>'
        '<S N="Line">10</S>'
        '<Nil N="MyCommand" />'
        '<I32 N="OffsetInLine">20</I32>'
        '<I32 N="PipelineLength">30</I32>'
        '<I32 N="PipelinePosition">40</I32>'
        '<S N="PositionMessage">position message</S>'
        '<Nil N="PSCommandPath" />'
        '<Nil N="PSScriptRoot" />'
        '<Nil N="ScriptLineNumber" />'
        '<Nil N="ScriptName" />'
        '<Obj RefId="5" N="UnboundArguments">'
        '<TN RefId="5">'
        "<T>System.Collections.ArrayList</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<LST>"
        "<B>true</B>"
        "</LST>"
        "</Obj>"
        "</Props>"
        "</Obj>"
        '<I32 N="ErrorCategory_Category">0</I32'
        '><Nil N="ErrorCategory_Activity" />'
        '<Nil N="ErrorCategory_Reason" />'
        '<Nil N="ErrorCategory_TargetName" />'
        '<Nil N="ErrorCategory_TargetType" />'
        '<S N="ErrorCategory_Message">NotSpecified (:) [], </S>'
        '<B N="SerializeExtendedInfo">false</B>'
        "</MS>"
        "<ToString>Exception</ToString>"
        "</Obj>"
    )
    assert_xml_diff(actual, expected)

    value.serialize_extended_info = True
    element = serialize(value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    expected = (
        '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Management.Automation.ErrorRecord</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<MS>"
        '<Obj RefId="1" N="Exception">'
        '<TN RefId="1">'
        "<T>System.Exception</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<Props>"
        '<S N="Message">Exception</S>'
        '<Nil N="Data" />'
        '<Nil N="HelpLink" />'
        '<Nil N="HResult" />'
        '<Nil N="InnerException" />'
        '<Nil N="Source" />'
        '<Nil N="StackTrace" />'
        '<Nil N="TargetSite" />'
        "</Props>"
        "</Obj>"
        '<Nil N="TargetObject" />'
        '<Nil N="FullyQualifiedErrorId" />'
        '<Obj RefId="2" N="InvocationInfo">'
        '<TN RefId="2">'
        "<T>System.Management.Automation.InvocationInfo</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<Props>"
        '<Obj RefId="3" N="BoundParameters">'
        '<TN RefId="3">'
        "<T>System.Collections.Hashtable</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<DCT>"
        '<En><S N="Key">Path</S><S N="Value">C:\\temp\\file.txt</S></En>'
        "</DCT>"
        "</Obj>"
        '<Obj RefId="4" N="CommandOrigin">'
        "<I32>0</I32>"
        '<TN RefId="4">'
        "<T>System.Management.Automation.CommandOrigin</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>Runspace</ToString>"
        "</Obj>"
        '<Nil N="DisplayScriptPosition" />'
        '<B N="ExpectingInput">false</B>'
        '<I64 N="HistoryId">10</I64>'
        '<S N="InvocationName">Remove-Item</S>'
        '<S N="Line">10</S>'
        '<Nil N="MyCommand" />'
        '<I32 N="OffsetInLine">20</I32>'
        '<I32 N="PipelineLength">30</I32>'
        '<I32 N="PipelinePosition">40</I32>'
        '<S N="PositionMessage">position message</S>'
        '<Nil N="PSCommandPath" />'
        '<Nil N="PSScriptRoot" />'
        '<Nil N="ScriptLineNumber" />'
        '<Nil N="ScriptName" />'
        '<Obj RefId="5" N="UnboundArguments">'
        '<TN RefId="5">'
        "<T>System.Collections.ArrayList</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<LST>"
        "<B>true</B>"
        "</LST>"
        "</Obj>"
        "</Props>"
        "</Obj>"
        '<I32 N="ErrorCategory_Category">0</I32>'
        '<Nil N="ErrorCategory_Activity" />'
        '<Nil N="ErrorCategory_Reason" />'
        '<Nil N="ErrorCategory_TargetName" />'
        '<Nil N="ErrorCategory_TargetType" />'
        '<S N="ErrorCategory_Message">NotSpecified (:) [], </S>'
        '<B N="SerializeExtendedInfo">true</B>'
        '<Obj RefId="6" N="InvocationInfo_BoundParameters">'
        '<TNRef RefId="3" />'
        "<DCT>"
        '<En><S N="Key">Path</S><S N="Value">C:\\temp\\file.txt</S></En>'
        "</DCT>"
        "</Obj>"
        '<Ref RefId="4" N="InvocationInfo_CommandOrigin" />'
        '<B N="InvocationInfo_ExpectingInput">false</B>'
        '<S N="InvocationInfo_InvocationName">Remove-Item</S>'
        '<S N="InvocationInfo_Line">10</S>'
        '<I32 N="InvocationInfo_OffsetInLine">20</I32>'
        '<I64 N="InvocationInfo_HistoryId">10</I64>'
        '<Obj RefId="7" N="InvocationInfo_PipelineIterationInfo">'
        '<TNRef RefId="5" />'
        "<LST />"
        "</Obj>"
        '<I32 N="InvocationInfo_PipelineLength">30</I32>'
        '<I32 N="InvocationInfo_PipelinePosition">40</I32>'
        '<Nil N="InvocationInfo_PSScriptRoot" />'
        '<Nil N="InvocationInfo_PSCommandPath" />'
        '<S N="InvocationInfo_PositionMessage">position message</S>'
        '<Nil N="InvocationInfo_ScriptLineNumber" />'
        '<Nil N="InvocationInfo_ScriptName" />'
        '<Obj RefId="8" N="InvocationInfo_UnboundArguments">'
        '<TNRef RefId="5" />'
        "<LST><B>true</B></LST>"
        '</Obj><B N="SerializeExtent">false</B>'
        '<Obj RefId="9" N="PipelineIterationInfo">'
        '<TNRef RefId="5" />'
        "<LST><I32>1</I32></LST>"
        "</Obj>"
        "</MS>"
        "<ToString>Exception</ToString>"
        "</Obj>"
    )
    assert_xml_diff(actual, expected)

    value = deserialize(element)

    assert isinstance(value, complex.ErrorRecord)
    assert str(value) == "Exception"
    assert value.serialize_extended_info is True
    assert value.Exception.Message == "Exception"

    # The exception contains the original invocation info and so doesn't have the re-computed values.
    assert isinstance(value.Exception.SerializedRemoteInvocationInfo, complex.InvocationInfo)
    assert value.Exception.SerializedRemoteInvocationInfo.PositionMessage == "position message"
    assert value.CategoryInfo.Category == complex.ErrorCategory.NotSpecified
    assert value.CategoryInfo.Activity is None
    assert value.CategoryInfo.Reason is None
    assert value.CategoryInfo.TargetName is None
    assert value.CategoryInfo.TargetType is None
    assert value.TargetObject is None
    assert value.FullyQualifiedErrorId is None
    assert value.InvocationInfo.BoundParameters == {"Path": "C:\\temp\\file.txt"}
    assert value.InvocationInfo.CommandOrigin == complex.CommandOrigin.Runspace

    display_script = value.InvocationInfo.DisplayScriptPosition
    assert isinstance(display_script, complex.ScriptExtent)
    assert display_script.StartOffset == 0
    assert display_script.StartLineNumber is None
    assert display_script.StartColumnNumber == 20
    assert display_script.EndOffset == 0
    assert display_script.EndLineNumber is None
    assert display_script.EndColumnNumber == 3
    assert display_script.File == ""
    assert display_script.Text == ""

    assert value.InvocationInfo.ExpectingInput is False
    assert value.InvocationInfo.HistoryId == 10
    assert value.InvocationInfo.InvocationName == "Remove-Item"
    assert value.InvocationInfo.Line == "10"
    assert value.InvocationInfo.MyCommand is None
    assert value.InvocationInfo.OffsetInLine == 20
    assert value.InvocationInfo.PSCommandPath == ""
    assert value.InvocationInfo.PSScriptRoot == ""
    assert value.InvocationInfo.PipelineLength == 30
    assert value.InvocationInfo.PipelinePosition == 40
    assert value.InvocationInfo.PositionMessage is None  # Haven't fully implemented these fields.
    assert value.InvocationInfo.ScriptLineNumber is None
    assert value.InvocationInfo.ScriptName is None
    assert value.InvocationInfo.UnboundArguments == [True]
    assert value.ErrorDetails is None
    assert value.PipelineIterationInfo == [1]
    assert value.ScriptStackTrace is None


def test_verbose_record_no_invocation_info():
    script_extent = complex.ScriptExtent(
        StartPosition=complex.ScriptPosition(
            File="my file",
            LineNumber=0,
            ColumnNumber=1,
            Line="my line",
        ),
        EndPosition=complex.ScriptPosition(
            File="my file",
            LineNumber=2,
            ColumnNumber=3,
            Line="my line",
        ),
    )
    invocation_info = complex.InvocationInfo(
        BoundParameters={"Param": "Value"},
        CommandOrigin=complex.CommandOrigin.Runspace,
        DisplayScriptPosition=script_extent,
        ExpectingInput=False,
        HistoryId=10,
        InvocationName="Invocation",
        Line=10,
        MyCommand=complex.PSCustomObject(
            CommandType=complex.CommandTypes.Function,
            Definition="command definition",
            Name="command name",
            Visibility=complex.SessionStateEntryVisibility.Private,
        ),
        OffsetInLine=2,
        PipelineLength=3,
        PipelinePosition=4,
        PSCommandPath="comand path",
        PSScriptRoot="script root",
        ScriptLineNumber=5,
        ScriptName="script name",
        UnboundArguments=["unbound", 1],
    )
    verbose = complex.VerboseRecord(
        Message="message",
        InvocationInfo=invocation_info,
        PipelineIterationInfo=[1, 2],
    )

    element = serialize(verbose)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    expected = (
        '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Management.Automation.VerboseRecord</T>"
        "<T>System.Management.Automation.InformationalRecord</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<MS>"
        '<S N="InformationalRecord_Message">message</S>'
        '<B N="InformationalRecord_SerializeInvocationInfo">false</B>'
        "</MS>"
        "</Obj>"
    )
    assert_xml_diff(actual, expected)

    verbose = deserialize(element)
    assert isinstance(verbose, complex.VerboseRecord)
    assert verbose.Message == "message"
    assert verbose.InvocationInfo is None
    assert verbose.PipelineIterationInfo is None
    assert not verbose.serialize_extended_info


def test_verbose_record_invocation_info():
    script_extent = complex.ScriptExtent(
        StartPosition=complex.ScriptPosition(
            File="my file",
            LineNumber=0,
            ColumnNumber=1,
            Line="my line",
        ),
        EndPosition=complex.ScriptPosition(
            File="my file",
            LineNumber=2,
            ColumnNumber=3,
            Line="my line",
        ),
    )
    invocation_info = complex.InvocationInfo(
        BoundParameters={"Param": "Value"},
        CommandOrigin=complex.CommandOrigin.Runspace,
        DisplayScriptPosition=script_extent,
        ExpectingInput=False,
        HistoryId=10,
        InvocationName="Invocation",
        Line=10,
        MyCommand=complex.PSCustomObject(
            CommandType=complex.CommandTypes.Function,
            Definition="command definition",
            Name="command name",
            Visibility=complex.SessionStateEntryVisibility.Private,
        ),
        OffsetInLine=2,
        PipelineLength=3,
        PipelinePosition=4,
        PSCommandPath="comand path",
        PSScriptRoot="script root",
        ScriptLineNumber=5,
        ScriptName="script name",
        UnboundArguments=["unbound", 1],
    )
    verbose = complex.VerboseRecord(
        Message="message",
        InvocationInfo=invocation_info,
        PipelineIterationInfo=[1, 2],
    )
    verbose.serialize_extended_info = True

    element = serialize(verbose)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    expected = (
        '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Management.Automation.VerboseRecord</T>"
        "<T>System.Management.Automation.InformationalRecord</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<MS>"
        '<S N="InformationalRecord_Message">message</S>'
        '<B N="InformationalRecord_SerializeInvocationInfo">true</B>'
        '<Obj RefId="1" N="InvocationInfo_BoundParameters">'
        '<TN RefId="1"><T>System.Collections.Hashtable</T><T>System.Object</T></TN>'
        '<DCT><En><S N="Key">Param</S><S N="Value">Value</S></En></DCT>'
        "</Obj>"
        '<Obj RefId="2" N="InvocationInfo_CommandOrigin">'
        "<I32>0</I32>"
        '<TN RefId="2">'
        "<T>System.Management.Automation.CommandOrigin</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN><ToString>Runspace</ToString>"
        "</Obj>"
        '<B N="InvocationInfo_ExpectingInput">false</B>'
        '<S N="InvocationInfo_InvocationName">Invocation</S>'
        '<S N="InvocationInfo_Line">10</S>'
        '<I32 N="InvocationInfo_OffsetInLine">2</I32>'
        '<I64 N="InvocationInfo_HistoryId">10</I64>'
        '<Obj RefId="3" N="InvocationInfo_PipelineIterationInfo">'
        '<TN RefId="3"><T>System.Collections.ArrayList</T><T>System.Object</T></TN>'
        "<LST />"
        '</Obj><I32 N="InvocationInfo_PipelineLength">3</I32>'
        '<I32 N="InvocationInfo_PipelinePosition">4</I32>'
        '<S N="InvocationInfo_PSScriptRoot">script root</S>'
        '<S N="InvocationInfo_PSCommandPath">comand path</S>'
        '<Nil N="InvocationInfo_PositionMessage" />'
        '<I32 N="InvocationInfo_ScriptLineNumber">5</I32>'
        '<S N="InvocationInfo_ScriptName">script name</S>'
        '<Obj RefId="4" N="InvocationInfo_UnboundArguments">'
        '<TNRef RefId="3" />'
        "<LST><S>unbound</S><I32>1</I32></LST>"
        "</Obj>"
        '<S N="ScriptExtent_File">my file</S>'
        '<I32 N="ScriptExtent_StartLineNumber">0</I32>'
        '<I32 N="ScriptExtent_StartColumnNumber">1</I32>'
        '<I32 N="ScriptExtent_EndLineNumber">2</I32>'
        '<I32 N="ScriptExtent_EndColumnNumber">3</I32>'
        '<B N="SerializeExtent">true</B>'
        '<Obj RefId="5" N="CommandInfo_CommandType">'
        "<I32>2</I32>"
        '<TN RefId="4">'
        "<T>System.Management.Automation.CommandTypes</T><T>System.Enum</T><T>System.ValueType</T><T>System.Object</T>"
        "</TN><ToString>Function</ToString>"
        "</Obj>"
        '<S N="CommandInfo_Definition">command definition</S>'
        '<S N="CommandInfo_Name">command name</S>'
        '<Obj RefId="6" N="CommandInfo_Visibility">'
        "<I32>1</I32>"
        '<TN RefId="5">'
        "<T>System.Management.Automation.SessionStateEntryVisibility</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN><ToString>Private</ToString>"
        "</Obj>"
        '<Obj RefId="7" N="InformationalRecord_PipelineIterationInfo">'
        '<TNRef RefId="3" />'
        "<LST><I32>1</I32><I32>2</I32></LST>"
        "</Obj>"
        "</MS>"
        "</Obj>"
    )
    assert_xml_diff(actual, expected)

    verbose = deserialize(element)
    assert isinstance(verbose, complex.VerboseRecord)
    assert verbose.Message == "message"

    invocation_info = verbose.InvocationInfo
    assert isinstance(invocation_info, complex.InvocationInfo)
    assert invocation_info.BoundParameters == {"Param": "Value"}
    assert invocation_info.CommandOrigin == complex.CommandOrigin.Runspace

    display_script = invocation_info.DisplayScriptPosition
    assert isinstance(display_script, complex.ScriptExtent)
    assert display_script.StartOffset == 0
    assert display_script.StartLineNumber == 0
    assert display_script.StartColumnNumber == 1
    assert display_script.EndOffset == 0
    assert display_script.EndLineNumber == 2
    assert display_script.EndColumnNumber == 3
    assert display_script.File == "my file"
    assert display_script.Text == "..."

    assert not invocation_info.ExpectingInput
    assert invocation_info.Line == "10"

    my_command = invocation_info.MyCommand
    assert isinstance(my_command, complex.RemoteCommandInfo)
    assert my_command.CommandType == complex.CommandTypes.Function
    assert my_command.Name == "command name"
    assert my_command.Definition == "command definition"
    assert my_command.Visibility == complex.SessionStateEntryVisibility.Private

    assert invocation_info.OffsetInLine == 2
    assert invocation_info.PipelineLength == 3
    assert invocation_info.PipelinePosition == 4
    assert invocation_info.PositionMessage is None
    assert invocation_info.PSCommandPath == ""
    assert invocation_info.PSScriptRoot == ""
    assert invocation_info.ScriptLineNumber == 5
    assert invocation_info.ScriptName == "script name"
    assert invocation_info.UnboundArguments == ["unbound", 1]

    assert verbose.PipelineIterationInfo == [1, 2]
    assert verbose.serialize_extended_info


def test_progress_record_defaults():
    record = complex.ProgressRecord(
        ActivityId=10,
        Activity=(COMPLEX_STRING + " - activity"),
        StatusDescription=(COMPLEX_STRING + " - status"),
    )
    element = serialize(record)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    expected = (
        "<PR>"
        f"<AV>{COMPLEX_ENCODED_STRING} - activity</AV>"
        "<AI>10</AI>"
        "<Nil />"
        "<PI>-1</PI>"
        "<PC>-1</PC>"
        "<T>Processing</T>"
        "<SR>-1</SR>"
        f"<SD>{COMPLEX_ENCODED_STRING} - status</SD>"
        "</PR>"
    )
    assert actual == expected

    record = deserialize(element)
    assert isinstance(record, complex.ProgressRecord)
    assert record.ActivityId == 10
    assert record.Activity == COMPLEX_STRING + " - activity"
    assert record.StatusDescription == COMPLEX_STRING + " - status"
    assert record.CurrentOperation is None
    assert record.ParentActivityId == -1
    assert record.PercentComplete == -1
    assert record.RecordType == complex.ProgressRecordType.Processing
    assert record.SecondsRemaining == -1


def test_progress_record_explicit():
    record = complex.ProgressRecord(
        ActivityId=10,
        Activity=(COMPLEX_STRING + " - activity"),
        StatusDescription=(COMPLEX_STRING + " - status"),
        CurrentOperation=(COMPLEX_STRING + " - operation"),
        ParentActivityId=20,
        PercentComplete=100,
        RecordType=complex.ProgressRecordType.Completed,
        SecondsRemaining=0,
    )
    element = serialize(record)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    expected = (
        "<PR>"
        f"<AV>{COMPLEX_ENCODED_STRING} - activity</AV>"
        "<AI>10</AI>"
        f"<S>{COMPLEX_ENCODED_STRING} - operation</S>"
        "<PI>20</PI>"
        "<PC>100</PC>"
        "<T>Completed</T>"
        "<SR>0</SR>"
        f"<SD>{COMPLEX_ENCODED_STRING} - status</SD>"
        "</PR>"
    )
    assert actual == expected

    record = deserialize(element)
    assert isinstance(record, complex.ProgressRecord)
    assert record.ActivityId == 10
    assert record.Activity == COMPLEX_STRING + " - activity"
    assert record.StatusDescription == COMPLEX_STRING + " - status"
    assert record.CurrentOperation == COMPLEX_STRING + " - operation"
    assert record.ParentActivityId == 20
    assert record.PercentComplete == 100
    assert record.RecordType == complex.ProgressRecordType.Completed
    assert record.SecondsRemaining == 0


def test_progress_record_missing_entry():
    data = "<PR><AV>activity</AV><AI>95</AI><PI /><SD>status</SD></PR>"
    record = deserialize(ElementTree.fromstring(data))

    assert isinstance(record, complex.ProgressRecord)
    assert record.ActivityId == 95
    assert record.Activity == "activity"
    assert record.StatusDescription == "status"
    assert record.CurrentOperation is None
    assert record.ParentActivityId == -1
    assert record.PercentComplete == -1
    assert record.RecordType == complex.ProgressRecordType.Processing
    assert record.SecondsRemaining == -1


def test_script_extent_text_no_end():
    extent = complex.ScriptExtent(
        StartPosition=complex.ScriptPosition(ColumnNumber=0), EndPosition=complex.ScriptPosition(ColumnNumber=0)
    )

    assert extent.Text == ""


def test_script_extent_text_same_line():
    extent = complex.ScriptExtent(
        StartPosition=complex.ScriptPosition(
            ColumnNumber=0,
            LineNumber=0,
            Line="test line",
        ),
        EndPosition=complex.ScriptPosition(
            ColumnNumber=6,
            LineNumber=0,
            Line="test line",
        ),
    )

    assert extent.Text == "test l"


def test_script_extent_text_multiple_lines():
    extent = complex.ScriptExtent(
        StartPosition=complex.ScriptPosition(
            ColumnNumber=2,
            LineNumber=0,
            Line="test line 2",
        ),
        EndPosition=complex.ScriptPosition(
            ColumnNumber=3,
            LineNumber=1,
            Line="test line 3",
        ),
    )

    assert extent.Text == "st line 2...t line 3"


def test_ps_primitive_dictionary():
    prim_dict = complex.PSPrimitiveDictionary(
        {
            "key": "value",
            "int key": 1,
            "casted": PSChar("a"),
        }
    )

    assert prim_dict["key"] == "value"
    assert prim_dict["int key"] == 1
    assert prim_dict["casted"] == 97
    assert isinstance(prim_dict["casted"], PSChar)

    element = serialize(prim_dict)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Management.Automation.PSPrimitiveDictionary</T>"
        "<T>System.Collections.Hashtable</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<DCT>"
        '<En><S N="Key">key</S><S N="Value">value</S></En>'
        '<En><S N="Key">int key</S><I32 N="Value">1</I32></En>'
        '<En><S N="Key">casted</S><C N="Value">97</C></En>'
        "</DCT>"
        "</Obj>"
    )

    prim_dict = deserialize(element)
    assert isinstance(prim_dict, complex.PSPrimitiveDictionary)
    assert isinstance(prim_dict, dict)
    assert prim_dict["key"] == "value"
    assert prim_dict["int key"] == 1
    assert prim_dict["casted"] == 97
    assert isinstance(prim_dict["casted"], PSChar)



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import uuid

from psrpcore import _events as events
from psrpcore.types import (
    PipelineState,
    PSInvocationState,
    PSRPMessageType,
    RunspaceAvailability,
    RunspacePoolState,
    RunspacePoolStateMsg,
)


def test_create_pipeline_state_event():
    msg = PipelineState(PipelineState=PSInvocationState.Completed)
    event = events.PSRPEvent.create(PSRPMessageType.PipelineState, msg, uuid.UUID(int=0), uuid.UUID(int=0))
    assert isinstance(event, events.PipelineStateEvent)
    assert event.state == PSInvocationState.Completed
    assert event.reason is None


def test_runspace_availability_set():
    msg = RunspaceAvailability(SetMinMaxRunspacesResponse=True, ci=1)
    event = events.PSRPEvent.create(PSRPMessageType.RunspaceAvailability, msg, uuid.UUID(int=0), None)
    assert isinstance(event, events.SetRunspaceAvailabilityEvent)
    assert event.success == True


def test_runspace_availability_get():
    msg = RunspaceAvailability(SetMinMaxRunspacesResponse=10, ci=1)
    event = events.PSRPEvent.create(PSRPMessageType.RunspaceAvailability, msg, uuid.UUID(int=0), None)
    assert isinstance(event, events.GetRunspaceAvailabilityEvent)
    assert event.count == 10


def test_runspace_pool_state_event():
    msg = RunspacePoolStateMsg(RunspaceState=RunspacePoolState.Opened)
    event = events.PSRPEvent.create(PSRPMessageType.RunspacePoolState, msg, uuid.UUID(int=0), None)
    assert isinstance(event, events.RunspacePoolStateEvent)
    assert event.state == RunspacePoolState.Opened
    assert event.reason is None



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from psrpcore import _crypto as crypto


def test_key_exchange():
    session_key = b"\x00" * 16

    private, public = crypto.create_keypair()
    assert isinstance(private, crypto.rsa.RSAPrivateKey)
    assert isinstance(public, bytes)
    assert private.key_size == 2048

    enc_session_key = crypto.encrypt_session_key(public, session_key)
    assert isinstance(enc_session_key, bytes)
    assert enc_session_key.startswith(b"\x01\x02\x00\x00\x10\x66\x00\x00\x00\xa4\x00\x00")

    actual_session_key = crypto.decrypt_session_key(private, enc_session_key)
    assert actual_session_key == session_key

    data = b"abc"
    encryptor = crypto.PSRemotingCrypto()
    encryptor.register_key(actual_session_key)
    enc_data = encryptor.encrypt(data)
    assert enc_data != data

    dec_data = encryptor.decrypt(enc_data)
    assert dec_data == data



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

"""Defines the base objects used for PSObjects.

This file contains the base classes and the various metadata/glue that is used
to represent a PSObject as a PowerShell class. It also contains the code
required to define a custom .NET type and add properties to those types that
is known by the serializer.

Also define some helper functions to replicate functionality in PowerShell/.NET
like `-is`, `Add-Member` and so on.
"""

import abc
import enum
import inspect
import types
import typing

T = typing.TypeVar("T", bound=typing.Type["PSObject"])


class _UnsetValue(object):
    """Used to mark a property with an unset value."""

    def __new__(  # type: ignore[misc]  # This is a sentinel value so it's expected
        cls,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> typing.Type["_UnsetValue"]:
        return cls  # pragma: no cover


class _Singleton(type):
    """Singleton used by TypeRegistry to ensure only 1 registry exists."""

    __instances: typing.Dict[typing.Type, object] = {}

    def __call__(cls, *args: typing.Any, **kwargs: typing.Any) -> object:
        if cls not in cls.__instances:
            cls.__instances[cls] = super().__call__(*args, **kwargs)

        return cls.__instances[cls]


class TypeRegistry(metaclass=_Singleton):
    """Registry of all the Python classes that implement PSObject.

    This singleton is used to store all the classes that implement PSObject
    and the .NET type it implements. This is used for deserialization to
    provide a dynamic list of Python classes that can be dehydrated.
    """

    def __init__(self) -> None:
        self.type_registry: typing.Dict[str, typing.Type["PSObject"]] = {}
        self.element_registry: typing.Dict[str, typing.Type["PSObject"]] = {}

    def rehydrate(
        self,
        type_names: typing.List[str],
    ) -> typing.Union["PSObject", typing.Type[enum.Enum]]:
        """Rehydrate a blank instance based on the type names."""
        obj: typing.Union[PSObject, typing.Type[enum.Enum]]

        type_name = type_names[0] if type_names else None
        if type_name and type_name in self.type_registry:
            # Cannot call __init__ as it may be validating input arguments which the serializer does not specify when
            # rehydrating that type.
            cls = self.type_registry[type_name]
            if issubclass(cls, enum.Enum):
                # The deserializer will rehydrate the object by providing the value.
                obj = cls

            else:
                obj = cls.__new__(cls)

        else:
            # The type is not registered, return a PSObject with the type names set to 'Deserialized.<TN>'.
            obj = PSObject()
            obj.PSObject.type_names = [f"Deserialized.{tn}" for tn in type_names]

        return obj


class PSObjectMeta:
    """The PowerShell PSObject metadata.

    This describes the metadata around the PSObject such as the properties and
    ETS info. This information is used by Python to (de)serialize the Python
    class to a .NET type through CLIXML. This is typically assigned internally
    by the `:class:PSType` class decorator and shouldn't be
    assigned directly.

    Using `rehydrate=True` (default) will register the type_name of the class
    so the deserializer will return an instance of that class when it comes to
    deserializing that type. A rehydrated object is created without calling
    __init__() so any validation or set up that occurs in that function when
    normally creating the class instance will not occur during deserialization
    and only the properties in CLIXML will be set on the class instance. When
    `rehydrate=False` then the deserialized object will be an instance of
    `class:PSObject` with the type names containing the `Deserialized.` prefix.

    Args:
        type_names: List of .NET type names that the type implements, this
            should contain at least 1 type.
        adapted_properties: List of adapted properties, these are native to
            the .NET type.
        extended_properties: List of extended properties, these are added to
            the .NET type by PowerShell.
        rehydrate: Whether the type should be registered as rehydratable or
            not.

    Attributes:
        type_names (List[str]): See args.
        adapted_properties (List[PSPropertyInfo]): See args.
        extended_properties (List[PSPropertyInfo]): See args.
        rehydrate (bool): See args.
    """

    def __init__(
        self,
        type_names: typing.List[str],
        adapted_properties: typing.Optional[typing.List["PSPropertyInfo"]] = None,
        extended_properties: typing.Optional[typing.List["PSPropertyInfo"]] = None,
        *,
        rehydrate: bool = True,
    ):
        self.type_names = type_names
        self.adapted_properties = adapted_properties or []
        self.extended_properties = extended_properties or []
        self.rehydrate = rehydrate

        self._to_string: typing.Optional[str] = None
        self._instance: typing.Optional[PSObject] = None

    @property
    def to_string(self) -> typing.Optional[str]:
        """The string representation of the object.

        The value to use for the `<ToString>` element of the serialized object.
        Will favour an explicit `to_string` value if set otherwise it will fall
        back to the value of `str(instance)` that the meta is for.
        """
        if self._instance is None or self._to_string is not None:
            return self._to_string

        # If the instance class of this object has an explicit __str__ method defined we use that as the to_string
        # value. We only want to check up to the PSObject() parent class in the mro because that falls back to this
        # property.
        for cls in type(self._instance).__mro__:
            if cls == PSObject:
                break

            if "__str__" in cls.__dict__:
                return str(self._instance)

        return None

    @to_string.setter
    def to_string(
        self,
        value: str,
    ) -> None:
        """Explicitly set the `to_string` value."""
        self._to_string = value

    def set_instance(
        self,
        instance: typing.Union["PSObject", typing.Type["PSObject"]],
    ) -> None:
        """Creates a copy of meta and assign to the class or instance."""
        kwargs: typing.Dict[str, typing.Any] = {
            "adapted_properties": [],
            "extended_properties": [],
        }
        for prop_name, prop_value in kwargs.items():
            for prop in getattr(self, prop_name):
                prop_value.append(prop.copy())

        kwargs["type_names"] = list(self.type_names)
        kwargs["rehydrate"] = self.rehydrate
        copy = type(self)(**kwargs)

        # If setting on an instance fo a PSObject, assign the instance to the copy.
        if isinstance(instance, PSObject):
            copy._instance = instance

        setattr(instance, "PSObject", copy)


class PSPropertyInfo(metaclass=abc.ABCMeta):
    """Base metadata for an object property.

    This is an abstract class that defines the property behaviour when it comes
    to getting and setting a property. There are three types of properties that
    have been implemented:

        :class:`PSAliasProperty`:
            A property that points to another property, or Python attribute.
            This essentially creates a getter that calls `ps_object['target']`.

        :class:`PSNoteProperty`:
            A property that contains it's own value like a normal
            attribute/property in Python.

        :class:`PSScriptProperty`:
            A property that uses a callable to get and optionally set a value
            on the ps_object.

    The `mandatory` option controls whether the default `__init__()` function
    added to PSObjects without their own `__init__()` will validate that
    property was specified by the caller. This has no control over the
    serialization behaviour and any classes that define their own `__init__()`
    need to do their own validation.

    Args:
        name: The name of the property.
        mandatory: The property must be defined when creating a PSObject.
        ps_type: If set, the property value will be casted to this PSObject
            type.
        value: The default value to set for the property.
        getter: A callable to get the property value based on the caller's
            desired logic. Must not be set with `value`.
        setter: A callable to set the property value based on the caller's
            desired logic. Must not be set with `value`.

    Attributes:
        name (str): See args.
        ps_type (type): See args.
        mandatory (bool): See args.
    """

    def __init__(
        self,
        name: str,
        mandatory: bool = False,
        ps_type: typing.Optional[type] = None,
        value: typing.Optional[typing.Any] = _UnsetValue,
        getter: typing.Optional[typing.Callable[["PSObject"], typing.Any]] = None,
        setter: typing.Optional[typing.Callable[["PSObject", typing.Any], None]] = None,
    ):
        self.name = name
        self.ps_type = ps_type
        self.mandatory = mandatory

        self._value: typing.Union[typing.Type[_UnsetValue], PSObject] = _UnsetValue

        self._getter: typing.Optional[typing.Callable[["PSObject"], typing.Any]] = None
        if getter:
            self.getter = getter

        self._setter: typing.Optional[typing.Callable[["PSObject", typing.Any], None]] = None
        if setter:
            self.setter = setter

        if value != _UnsetValue:
            if getter:
                raise ValueError(f"Cannot set property value for '{self.name}' with a getter")

            # The PSObject is required when setting a value for a custom setter. Because we do not set a value if there
            # is a getter/setter present then this can be a random object without causing any issues.
            self.set_value(value, PSObject())

    @abc.abstractmethod
    def copy(self) -> "PSPropertyInfo":
        """Create a copy of the property."""
        pass  # pragma: no cover

    @property
    def getter(self) -> typing.Optional[typing.Callable[["PSObject"], typing.Any]]:
        """Returns the getter callable for the property if one was set."""
        return self._getter

    @getter.setter
    def getter(
        self,
        getter: typing.Callable[["PSObject"], None],
    ) -> None:
        if self._value != _UnsetValue:
            raise ValueError(f"Cannot add getter for '{self.name}': existing value already set")

        if getter is None:
            raise ValueError(f"Cannot unset property getter for '{self.name}'")

        self._validate_callable(getter, 1, "getter")
        self._getter = getter

    @property
    def setter(self) -> typing.Optional[typing.Callable[["PSObject", typing.Any], None]]:
        """Returns the setter callable for the property if one was set."""
        return self._setter

    @setter.setter
    def setter(
        self,
        setter: typing.Optional[typing.Callable[["PSObject", typing.Any], None]],
    ) -> None:
        if self.getter is None:
            raise ValueError(f"Cannot set property setter for '{self.name}' without an existing getter")

        elif setter is None:
            self._setter = None

        else:
            self._validate_callable(setter, 2, "setter")
            self._setter = setter

    def get_value(
        self,
        ps_object: "PSObject",
    ) -> typing.Any:
        """Get the property value.

        Gets the value of the property. If the property value is a callable
        then the value is invoked with the ps_object as an argument and the
        resulting value is casted to the `ps_type` if it is set.

        Args:
            ps_object: The PSObject instance the property is on. This is used
                when invoking a getter callable.

        Returns:
            (typing.Any): The value of the property.
        """
        getter = self.getter
        value: typing.Optional[typing.Union[typing.Type[_UnsetValue], PSObject]]

        if getter:
            raw_value = getter(ps_object)
            value = self._cast(raw_value)

        else:
            value = self._value

        if value == _UnsetValue:
            value = None

        return value

    def set_value(
        self,
        value: typing.Any,
        ps_object: "PSObject",
    ) -> None:
        """Set the property value.

        Sets the value of the property to the value specified. The value will
        be casted to the property's `ps_type` if defined unless it is a
        callable or `None`. Trying to set `None` on a `mandatory` property will
        also fail.

        Args:
            value: The value to set on the property.
            ps_object: The PSObject instance the property is on. This is used
                when invoking the setter callable.
        """
        setter = self.setter
        if setter:
            setter(ps_object, value)

        elif self.getter:
            raise ValueError(f"Cannot set value for a getter property '{self.name}' without a setter callable")

        else:
            self._value = self._cast(value)

    def _cast(self, value: typing.Any) -> "PSObject":
        """Try to cast the raw value to the property's ps_type if possible."""
        if value is not None and self.ps_type is not None and not isinstance(value, self.ps_type):
            return self.ps_type(value)

        else:
            return value

    def _validate_callable(
        self,
        func: typing.Callable,
        expected_count: int,
        use: str,
    ) -> None:
        """Validates the callable has the required argument count for use as a property getter/setter."""
        if not isinstance(func, types.FunctionType):
            raise TypeError(
                f"Invalid {use} callable for property '{self.name}': expecting callable not "
                f"{type(func).__qualname__}"
            )

        parameters = list(inspect.signature(func).parameters.values())
        required_count = 0
        total_count = 0

        for param in parameters:
            if param.kind in [inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD]:
                # func(arg1, .., /) or func(arg1, ...) - keep track of how many and if they must be set.
                total_count += 1
                if param.default == inspect.Parameter.empty:
                    required_count += 1

            elif param.kind == inspect.Parameter.VAR_POSITIONAL:
                # Once we've reached *args we've counted all the positional args that could be used. It also means the
                # callable accepts an arbitrary amount of args so our expected count will be met.
                total_count = expected_count
                break

        def plural(name: str, count: int) -> str:
            s = "" if count == 1 else "s"
            return f"{count} {name}{s}"

        base_err = (
            f"Invalid {use} callable for property '{self.name}': signature expected "
            f"{plural('parameter', expected_count)} but"
        )
        if required_count > expected_count:
            raise TypeError(f"{base_err} {plural('required parameter', required_count)} were found")

        elif total_count < expected_count:
            raise TypeError(f"{base_err} {plural('parameter', total_count)} were found")


class PSAliasProperty(PSPropertyInfo):
    """Alias Property

    This is a property that gets a value based on another property or attribute
    of the PSObject. It is designed to replicate the functionality of
    `PSAliasProperty`_. During serialization the alias property will just copy
    the value of the target it is pointing to. You cannot set a value to an
    alias property, see :class:`PSScriptProperty` which allows the caller
    to define a way to get and set properties on an object dynamically.

    Note:
        When an object that has an alias property is deserialized, the property
        is converted to a :class:`PSNoteProperty` and the alias will no longer
        exist.

    Args:
        name: The name of the property.
        alias: The name of the property or attribute to point to.
        ps_type: If set, the property value will be casted to this PSObject
            type.

    Attributes:
        alias (str): The target of the alias.

    .. _PSAliasProperty:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.psaliasproperty
    """

    def __init__(
        self,
        name: str,
        alias: str,
        ps_type: typing.Optional[type] = None,
    ):
        self.alias = alias
        super().__init__(name, ps_type=ps_type, getter=lambda s: s[alias])

    def copy(self) -> "PSAliasProperty":
        return PSAliasProperty(self.name, self.alias, self.ps_type)


class PSNoteProperty(PSPropertyInfo):
    """Note Property

    This is a property that stores a value as a name-value pair. Is is designed
    to replicate the functionality of `PSNoteProperty`_ and is typically the
    type of property to use when creating a PSObject.

    Note:
        See :class:`PSPropertyInfo` for more information on the `mandatory`
        argument.

    Args:
        name: The name of the property.
        value: The property value to set, if omitted the default is `None`.
        mandatory: The property must be defined when creating a PSObject.
        ps_type: If set, the property value will be casted to this PSObject type.

    .. _PSNoteProperty:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.psnoteproperty
    """

    def __init__(
        self,
        name: str,
        value: typing.Optional[typing.Any] = _UnsetValue,
        mandatory: bool = False,
        ps_type: typing.Optional[type] = None,
    ):
        super().__init__(name, mandatory=mandatory, ps_type=ps_type, value=value)

    def copy(self) -> "PSNoteProperty":
        return PSNoteProperty(self.name, self._value, self.mandatory, self.ps_type)


class PSScriptProperty(PSPropertyInfo):
    """Script Property

    This is a property that can get and optionally set another property or
    attribute of a PSObject at runtime. It is designed to replicate the
    functionality of `PSScriptProperty`_.

    The getter callable must be a callable that has only 1 argument that is the
    PSObject the property is a member of. This allows the caller to retrieve a
    property of the PSObject at runtime or any other source as needed.

    The setter callable must be a callable that has only 2 arguments, the first
    being the value that needs to be set and the second is the PSObject the
    property is a member of. A setter must be defined on the property for a
    value to be set.

    Args:
        name: The name of the property.
        getter: The callable to run when getting a value for this property.
        setter: The callable to run when setting a value for this property.
        mandatory: The property must be defined when creating a PSObject.
        ps_type: If set, the property value will be casted to this PSObjec
             type.

    .. _PSScriptProperty:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.psscriptproperty
    """

    def __init__(
        self,
        name: str,
        getter: typing.Callable[["PSObject"], typing.Any],
        setter: typing.Optional[typing.Callable[["PSObject", typing.Any], None]] = None,
        mandatory: bool = False,
        ps_type: typing.Optional[type] = None,
    ):
        if getter is None:
            raise TypeError(f"Cannot create script property '{name}' with getter as None")

        if mandatory and not setter:
            raise TypeError(
                f"Cannot create mandatory {self.__class__.__qualname__} property '{name}' without a " f"setter callable"
            )

        super().__init__(name, mandatory=mandatory, ps_type=ps_type, getter=getter, setter=setter)

    def copy(self) -> "PSScriptProperty":
        return PSScriptProperty(
            self.name,
            self.getter,  # type: ignore[arg-type] # The class does not allow self.getter to be None
            self.setter,
            self.mandatory,
            self.ps_type,
        )


class PSObject:
    """The base PSObject type.

    This is the base PSObject type that all PS object classes must inherit. It
    controls all the behaviour around getting and setting attributes that are
    based on PowerShell properties in a way that is similar to PowerShell
    itself.

    When initialised directly this is a plain object without any properties.
    User :meth:`add_member` to add new properties dynamically.
    """

    PSObject = PSObjectMeta(
        type_names=[
            "System.Object",
        ]
    )  #: Class and instance attribute that defines the objects ETS.

    def __new__(  # type: ignore[no-untyped-def]  # Trying to set PSObject here sends mypy into haywire
        cls,
        *args: typing.Any,
        **kwargs: typing.Any,
    ):
        if super().__new__ is object.__new__ and cls.__init__ is not object.__init__:
            obj = super().__new__(cls)
        else:
            obj = super().__new__(cls, *args, **kwargs)

        # Make sure the class instance has a copy of the class PSObject so they aren't shared. Also add a reference to
        # the instance for that PSObject.
        cls.PSObject.set_instance(obj)

        return obj

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        # Skip creating a new object if we are trying to cast to the same type again.
        # if len(args) == 1 and type(args[0]) == cls:
        #    return args[0]

        # Favour extended properties in case there is one with the same name. This is how PowerShell's ETS works.
        prop_entries = {p.name: p for p in self.PSObject.adapted_properties}
        prop_entries.update({p.name: p for p in self.PSObject.extended_properties})

        # Make sure the number of positional args specified do not exceed the number of kwargs present.
        if len(args) > len(prop_entries):
            raise TypeError(
                f"__init__() takes {len(prop_entries) + 1} positional arguments but {len(args) + 1} " f"were given"
            )

        # Convert the args to kwargs based on the property order and check that they aren't also defined as a kwarg.
        caller_args = dict(zip(prop_entries.keys(), args))
        for name in caller_args.keys():
            if name in kwargs:
                raise TypeError(f"__init__() got multiple values for argument '{name}'")
        caller_args.update(kwargs)

        # Validate that any mandatory props were specified.
        # Cannot use set as it breaks the ordering which we want to preserve for the error msg.
        mandatory_props = [p.name for p in prop_entries.values() if p.mandatory]
        specified_props = list(caller_args.keys())
        missing_props = [p for p in mandatory_props if p not in specified_props]
        if missing_props:
            missing_list = "', '".join(missing_props)
            raise TypeError(f"__init__() missing {len(missing_props)} required arguments: '{missing_list}'")

        # Check that all the kwargs match the existing properties of the object and set the properties.
        for prop_name, raw_value in caller_args.items():
            if prop_name not in prop_entries:
                raise TypeError(f"__init__() got an unexpected keyword argument '{prop_name}'")

            setattr(self, prop_name, raw_value)

    @property
    def PSBase(self) -> None:
        """The raw .NET object without the extended type system properties."""
        raise NotImplementedError()  # pragma: no cover

    @property
    def PSAdapted(self) -> None:
        """A dict of all the adapted properties."""
        raise NotImplementedError()  # pragma: no cover

    @property
    def PSExtended(self) -> None:
        """A dict of all the extended properties."""
        raise NotImplementedError()  # pragma: no cover

    @property
    def PSTypeNames(self) -> typing.List[str]:
        """Shortcut to PSObject.type_names, one of PowerShell's reserved properties."""
        return self.PSObject.type_names

    def __setattr__(self, key: str, value: typing.Any) -> None:
        # __getattribute__ uses PSObject so bypass all that and just set it directly.
        if key == "PSObject":
            return super().__setattr__(key, value)

        # Get the raw untainted __dict__ value which does not include our object's PS properties that self.__dict__
        # will return. We use this to see whether we need to set the PSObject property or the Python object attribute.
        d = super().__getattribute__("__dict__")

        if key not in d:
            ps_object = self.PSObject

            # Extended props take priority, once we find a match we stopped checking.
            for prop_type in ["extended", "adapted"]:
                properties = getattr(ps_object, f"{prop_type}_properties")
                for prop in properties:
                    if prop.name == key:
                        prop.set_value(value, self)
                        return

        # If the key already exists in the __dict__ or it's a new attribute that's not a registered property, just
        # set the key/value to the object itself.
        super().__setattr__(key, value)

    def __getattribute__(self, item: str) -> typing.Any:
        # Use __getattribute__ instead of self.PSObject to avoid a recursive call.
        ps_object = super().__getattribute__("PSObject")

        # Extended props take priority over adapted props, by checking that last we ensure the prop will have that
        # value if there are duplicates.
        ps_properties = {}
        for prop_type in ["adapted", "extended"]:
            properties = getattr(ps_object, f"{prop_type}_properties")
            for prop in properties:
                ps_properties[prop.name] = prop

        # We want to favour the normal attributes Python would return before falling back to the properties on the
        # PSObject.
        try:
            val = super().__getattribute__(item)
        except AttributeError:
            if item not in ps_properties:
                raise

            return ps_properties[item].get_value(self)

        # A special case exists when returning __dict__. We want to have __dict__ return both the Python attributes as
        # well as the PSObject properties. This allows debuggers and people calling functions like vars()/dirs() to see
        # the PSObject properties automatically.
        if item == "__dict__":
            val = val.copy()  # Make sure we don't actually mutate the pure __dict__.
            val.update({name: prop.get_value(self) for name, prop in ps_properties.items()})

        return val

    def __getitem__(self, item: str) -> typing.Any:
        """Allow getting properties using the index syntax.

        By overriding __getitem__ you can access properties on an object using
        the index syntax, i.e. obj['PropertyName']. This matches the PowerShell
        behaviour where properties can be retrieved either by dot notation or
        by index notation.

        It also makes it easier to get properties with a name that aren't valid
        Python attribute names. By allowing a string field someone can do
        `obj['1 Invalid Attribute$']`. An alternative option is through
        getattr() as that accepts a string. This works because PSObject also
        override :func:`__getattr__` and :func:`__setattr__` and it edits the
        `__dict__` directly.

        This is complicated by the Dict/List/Stack/Queue types as we need this
        to preserve the actual lookup values. In those cases the __getitem__
        lookup will favour the base object items before falling back to looking
        at the attributes.
        """
        return getattr(self, item)

    def __setitem__(self, key: str, value: typing.Any) -> None:
        setattr(self, key, value)

    def __str__(self) -> str:
        if self.PSObject.to_string is not None:
            return self.PSObject.to_string

        else:
            return super().__str__()

    def __repr__(self) -> str:
        prop_entries = {p.name: p for p in self.PSObject.adapted_properties}
        prop_entries.update({p.name: p for p in self.PSObject.extended_properties})

        kw = ", ".join(f"{k}={v.get_value(self)!r}" for k, v in prop_entries.items())
        return f"{type(self).__name__}({kw})"


class PSType:
    """PSType class decorator.

    This is a class decorator that should be added to any Python class that
    represents a .NET type. It modifies the class to use the PowerShell type
    system like extended properties. A class that uses this decorator must
    inherit from :class:`PSObject`.

    If the class also has a base `PSType` class it will inherit all the types
    and properties of that base class unless `skip_inheritance=True` is set.

    Args:
        type_names: A list of .NET types the class represents.
        adapted_properties: A list of :class:`PSPropertyInfo` to denote the
            native .NET properties the class has.
        extended_properties: A list of :class:`PSPropertyInfo` to denote the
            extended PowerShell properties the class has.
        skip_inheritance: Skip inheriting the type and properties from the base
            class.
        rehydrate: Registers the type for deserialisation.
        tag: Used internally to denote the CLIXML element tag value.
    """

    def __init__(
        self,
        type_names: typing.List[str] = None,
        adapted_properties: typing.List[PSPropertyInfo] = None,
        extended_properties: typing.List[PSPropertyInfo] = None,
        *,
        skip_inheritance: bool = False,
        rehydrate: bool = True,
        tag: typing.Optional[str] = None,
    ):
        self.type_names = type_names or []
        self.adapted_properties = adapted_properties or []
        self.extended_properties = extended_properties or []
        self.skip_inheritance = skip_inheritance
        self.rehydrate = rehydrate
        self.tag = tag

    def __call__(
        self,
        cls: T,
    ) -> T:
        if not issubclass(cls, PSObject):
            raise TypeError(f"PSType class {cls.__module__}.{cls.__qualname__} must be a subclass of PSObject")

        if not self.skip_inheritance:
            self.type_names.extend(cls.PSObject.type_names)
            self.adapted_properties.extend(cls.PSObject.adapted_properties)
            self.extended_properties.extend(cls.PSObject.extended_properties)

        cls.PSObject = PSObjectMeta(self.type_names, self.adapted_properties, self.extended_properties)

        registry = TypeRegistry()

        if self.rehydrate and self.type_names and self.type_names[0] not in registry.type_registry:
            registry.type_registry[self.type_names[0]] = cls

        if self.tag is not None and self.tag not in registry.element_registry:
            registry.element_registry[self.tag] = cls

        return cls


class PSCryptoProvider(metaclass=abc.ABCMeta):
    """PSRemoting crypto provider

    The CryptoProvider used by PSRemoting that can encrypt and decrypt secure
    exchanged in that PSSession. The key must be registered once it has been
    generated using :meth:`PSCryptoProvider.register_key`.
    """

    @abc.abstractmethod
    def decrypt(self, value: bytes) -> bytes:
        """Decrypts the encrypted bytes.

        Decrypts the encrypted bytes passed in.

        Args:
            value: The encrypted bytes to decrypt.

        Returns:
            bytes: The decrypted bytes.
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def encrypt(self, value: bytes) -> bytes:
        """Encrypted the bytes.

        Encrypts the bytes passed in.

        Args:
            value: The bytes to encrypt.

        Returns:
            bytes: The encrypted bytes.
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def register_key(
        self,
        key: bytes,
    ) -> None:
        """Registers the session key.

        Registers the session key that is used to encrypt and decrypt secure
        strings for PSRP.

        Args:
            key: The session key negotiated between the client and server.
        """
        pass  # pragma: no cover


def add_member(
    obj: typing.Union[PSObject, typing.Type[PSObject]],
    prop: PSPropertyInfo,
    force: bool = False,
) -> None:
    """Add an extended property.

    This can add an extended property to a PSObject class or a specific
    instance of a class. This replicates some of the functionality in
    `Update-TypeData`_ and `Add-Member`_ in PowerShell. If a property under the
    same name already exists under that PSObject then `force=True` is required
    to replace it. The same applies if there is an existing adapted property on
    the object.

    See :meth:`add_alias_property`, :meth:`add_note_property`, and
    :meth:`add_script_property` for simplified versions of this function.

    Args:
        obj: The PSObject class or an instance of a PSObject class to add the
            extended property to.
        prop: The property to add to the object or class.
        force: Overwrite the existing property ``True`` or fail ``False``.

    .. _Update-TypeData:
        https://docs.microsoft.com/en-us/powershell/module/microsoft.powershell.utility/update-typedata
    .. _Add-Member:
        https://docs.microsoft.com/en-us/powershell/module/microsoft.powershell.utility/add-member
    """
    ps_object = getattr(obj, "PSObject", None)
    if not ps_object:
        raise ValueError("The passing in object does not contain the required PSObject attribute")

    prop_name = prop.name
    adapted_properties = {p.name: i for i, p in enumerate(ps_object.adapted_properties)}
    extended_properties = {p.name: i for i, p in enumerate(ps_object.extended_properties)}

    insert_idx = len(ps_object.extended_properties)
    if prop_name in adapted_properties or prop_name in extended_properties:
        if not force:
            raise RuntimeError(f"Property '{prop_name}' already exists on PSObject, use force=True to overwrite it")

        # If we had a duplicated extended prop, swap the older with the new one, adapted props stays the same.
        if prop_name in extended_properties:
            insert_idx = extended_properties[prop_name]
            ps_object.extended_properties.pop(insert_idx)

    ps_object.extended_properties.insert(insert_idx, prop)


def add_alias_property(
    obj: typing.Union[PSObject, typing.Type[PSObject]],
    name: str,
    alias: str,
    ps_type: typing.Optional[type] = None,
    force: bool = False,
) -> None:
    """Add an alias property to a PSObject.

    Adds an alias as an extended property to a PSObject class or a specific
    instance of a class.

    Args:
        obj: The PSObject to add the alias to.
        name: The name of the new alias property.
        alias: The alias target.
        ps_type: Optional PSObject type that the alias value will get casted to.
        force: Overwrite the existing property ``True`` or fail ``False``.
    """
    add_member(obj, PSAliasProperty(name, alias, ps_type=ps_type), force=force)


def add_note_property(
    obj: typing.Union[PSObject, typing.Type[PSObject]],
    name: str,
    value: typing.Any,
    ps_type: typing.Optional[type] = None,
    force: bool = False,
) -> None:
    """Add a note property to a PSObject.

    Adds a note property as an extended property to a PSObject class or a
    specific instance of a class. A note property is a simple key/value pair
    with a static value.

    Args:
        obj: The PSObject to add the note property to.
        name: The name of the new note property.
        value: The value of the new note property
        ps_type: Optional PSObject type that the value will get casted to.
        force: Overwrite the existing property ``True`` or fail ``False``.
    """
    add_member(obj, PSNoteProperty(name, value, ps_type=ps_type), force=force)


def add_script_property(
    obj: typing.Union[PSObject, typing.Type[PSObject]],
    name: str,
    getter: typing.Callable[["PSObject"], typing.Any],
    setter: typing.Optional[typing.Callable[["PSObject", typing.Any], None]] = None,
    ps_type: typing.Optional[type] = None,
    force: bool = False,
) -> None:
    """Add a script property to a PSObject.

    Adds a script property as an extended property to a PSObject class or a
    specific instance of a class. A script property has a callable getter and
    optional setter function that is run when the property's value is requested
    or set.

    Args:
        obj: The PSObject to add the alias to.
        name: The name of the new alias property.
        getter: The callable to run when getting a value for this property.
        setter: The callable to run when setting a value for this property.
        ps_type: Optional PSObject type that the alias value will get casted to.
        force: Overwrite the existing property ``True`` or fail ``False``.
    """
    add_member(obj, PSScriptProperty(name, getter, setter=setter, ps_type=ps_type), force=force)


def ps_isinstance(
    obj: PSObject,
    other: typing.Union[typing.Type[PSObject], typing.Tuple[typing.Type[PSObject], ...], str, typing.Tuple[str, ...]],
    ignore_deserialized: bool = False,
) -> bool:
    """Checks the inheritance of a PSObject.

    This checks if a PSObject is an instance of another PSObject. Instead of
    checking based on the Python inheritance rules it checks based on the
    .NET TypeNames set for that instance. The check will loop through the
    `PSTypeNames` of the obj passed in and see if any of those types match the
    first `PSTypeName` of any of the `other` objects referenced.

    If `check_deserialized=True`, then any types starting with `Deserialized.`
    will also match against the non-deserialized types, e.g.
    `Deserialized.System.Collections.Hashtable` will be an instance of
    `System.Collections.Hashtable`.

    Args:
        obj: The object to check if it is inherited from the other types.
        other: The type to check if obj is inherited from. Can also be a list
            of .NET types as a string.
        ignore_deserialized: Whether to treat `Deserialized.*` instances as
            they would be serialized.

    Returns:
        bool: Whether the obj is inherited from any of the other types in .NET.
    """

    def strip_deserialized(type_names: typing.List[str]) -> typing.List[str]:
        if not ignore_deserialized:
            return type_names

        new_names = []
        for name in type_names:
            if name.startswith("Deserialized."):
                name = name[13:]

            new_names.append(name)

        return new_names

    other_instances = other if isinstance(other, (list, tuple)) else [other]
    raw_other_types = []
    for o in other_instances:
        if isinstance(o, str):
            raw_other_types.append(o)
        elif hasattr(o, "PSObject"):
            raw_other_types.append(o.PSObject.type_names[0])

    obj_types = set(strip_deserialized(obj.PSTypeNames))
    desired_types = set(strip_deserialized(raw_other_types))
    matching_types = obj_types & desired_types

    return bool(matching_types)



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from psrpcore._payload import ProtocolVersion, dict_to_psobject
from psrpcore.types import PipelineResultTypes, PSObject, PSType


@PSType(skip_inheritance=True, rehydrate=False)
class Command(PSObject):
    """Pipeline Command.

    Defines a Command object which can be added to a Pipeline for invocation.

    Args:
        name: The command name or script contents.
        is_script: If the command represents a script or command.
        use_local_scope: The command is run in local scope or not.

    Attributes:
        command_text: The command name or script.
        is_script: See args.
        use_local_scope: See args.
        parameters: List of parameters to invoke with the command.
        end_of_statement: Whether the command is the last one in the current
            statement or not.
        merge_unclaimed: Set as the merge point for any error records generated
            by previous commands in the statement.
    """

    def __init__(
        self,
        name: str,
        is_script: bool = False,
        use_local_scope: typing.Optional[bool] = None,
    ) -> None:
        self.command_text = name
        self.is_script = is_script
        self.use_local_scope = use_local_scope
        self.parameters: typing.List[typing.Tuple[typing.Optional[str], typing.Any]] = []
        self.end_of_statement = False

        self.merge_unclaimed = False
        self._merge_my = PipelineResultTypes.none
        self._merge_to = PipelineResultTypes.none
        self._merge_error = PipelineResultTypes.none
        self._merge_warning = PipelineResultTypes.none
        self._merge_verbose = PipelineResultTypes.none
        self._merge_debug = PipelineResultTypes.none
        self._merge_information = PipelineResultTypes.none

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__} command_text={self.command_text!r} is_script={self.is_script} "
            f"use_local_scope={self.use_local_scope} end_of_statement={self.end_of_statement}>"
        )

    def __str__(self) -> str:
        return self.command_text

    @property
    def merge_my(self) -> PipelineResultTypes:
        """Used for v2 compatibility to merge error to output/null."""
        return self._merge_my

    @property
    def merge_to(self) -> PipelineResultTypes:
        """Used for v2 compatibility to merge error to output/null."""
        return self._merge_to

    @property
    def merge_error(self) -> PipelineResultTypes:
        """What stream to send error records to."""
        return self._merge_error

    @property
    def merge_warning(self) -> PipelineResultTypes:
        """What stream to send warning records to."""
        return self._merge_warning

    @property
    def merge_verbose(self) -> PipelineResultTypes:
        """What stream to send verbose records to."""
        return self._merge_verbose

    @property
    def merge_debug(self) -> PipelineResultTypes:
        """What stream to send debug records to."""
        return self._merge_debug

    @property
    def merge_information(self) -> PipelineResultTypes:
        """What stream to send information records to."""
        return self._merge_information

    def add_argument(
        self,
        value: typing.Any,
    ) -> "Command":
        """Add a positional argument to the command."""
        return self.add_parameter(None, value)

    def add_parameter(
        self,
        name: typing.Optional[str],
        value: typing.Any = True,
    ) -> "Command":
        """Add a parameter and value to the command."""
        self.parameters.append((name, value))
        return self

    def add_parameters(
        self,
        **parameters: typing.Any,
    ) -> "Command":
        """Adds mutliple parameters to the command."""
        for name, value in parameters.items():
            self.add_parameter(name, value)

        return self

    def redirect_all(
        self,
        stream: PipelineResultTypes = PipelineResultTypes.Output,
    ) -> None:
        """Redirects all streams.

        Redirects all the non-output streams to the targeted stream. The target
        stream can only be ``none``, ``Output``, or ``Null``.

        Args:
            stream: The stream to redirect to.
        """
        self.redirect_error(stream)
        self.redirect_warning(stream)
        self.redirect_verbose(stream)
        self.redirect_debug(stream)
        self.redirect_information(stream)

    def redirect_error(
        self,
        stream: PipelineResultTypes = PipelineResultTypes.Output,
    ) -> None:
        """Redirect the error stream to this."""
        self._validate_redirection_to(stream)
        if stream == PipelineResultTypes.none:
            self._merge_my = PipelineResultTypes.none
            self._merge_to = PipelineResultTypes.none

        elif stream != PipelineResultTypes.Null:
            self._merge_my = PipelineResultTypes.Error
            self._merge_to = stream

        self._merge_error = stream

    def redirect_warning(
        self,
        stream: PipelineResultTypes = PipelineResultTypes.Output,
    ) -> None:
        """Redirect the warning stream to this."""
        self._validate_redirection_to(stream)
        self._merge_warning = stream

    def redirect_verbose(
        self,
        stream: PipelineResultTypes = PipelineResultTypes.Output,
    ) -> None:
        """Redirect the verbose stream to this."""
        self._validate_redirection_to(stream)
        self._merge_verbose = stream

    def redirect_debug(
        self,
        stream: PipelineResultTypes = PipelineResultTypes.Output,
    ) -> None:
        """Redirect the debug stream to this."""
        self._validate_redirection_to(stream)
        self._merge_debug = stream

    def redirect_information(
        self,
        stream: PipelineResultTypes = PipelineResultTypes.Output,
    ) -> None:
        """Redirect the information stream to this."""
        self._validate_redirection_to(stream)
        self._merge_information = stream

    def _validate_redirection_to(
        self,
        stream: PipelineResultTypes,
    ) -> None:
        if stream not in [PipelineResultTypes.none, PipelineResultTypes.Output, PipelineResultTypes.Null]:
            raise ValueError("Invalid redirection stream, must be none, Output, or Null")

    @classmethod
    def FromPSObjectForRemoting(
        cls,
        obj: PSObject,
        **kwargs: typing.Any,
    ) -> "Command":
        cmd = cls(
            name=obj.Cmd,
            is_script=obj.IsScript,
            use_local_scope=obj.UseLocalScope,
        )
        for argument in obj.Args:
            cmd.parameters.append((argument.N, argument.V))

        merge_unclaimed = PipelineResultTypes.Output | PipelineResultTypes.Error
        cmd.merge_unclaimed = bool(obj.MergePreviousResults == merge_unclaimed)

        cmd._merge_my = obj.MergeMyResult
        cmd._merge_to = obj.MergeToResult

        # Depending on the peer protocolversion, these fields may not be present.
        for name in ["Error", "Warning", "Verbose", "Debug", "Information"]:
            value = getattr(obj, f"Merge{name}", None)
            if value is not None:
                setattr(cmd, f"_merge_{name.lower()}", value)

        return cmd

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "Command",
        **kwargs: typing.Any,
    ) -> PSObject:
        protocol_version = None
        if "their_capability" in kwargs:
            protocol_version = getattr(kwargs["their_capability"], "protocolversion", None)

        if not protocol_version:
            protocol_version = ProtocolVersion.Win7RC.value

        merge_previous = (
            PipelineResultTypes.Output | PipelineResultTypes.Error
            if instance.merge_unclaimed
            else PipelineResultTypes.none
        )

        command_kwargs = {
            "Cmd": instance.command_text,
            "Args": [dict_to_psobject(N=n, V=v) for n, v in instance.parameters],
            "IsScript": instance.is_script,
            "UseLocalScope": instance.use_local_scope,
            "MergeMyResult": instance.merge_my,
            "MergeToResult": instance.merge_to,
            "MergePreviousResults": merge_previous,
        }

        # For backwards compatibility we need to optional set these values based on the peer's protocol version.
        if protocol_version >= ProtocolVersion.Pwsh3.value:
            command_kwargs["MergeError"] = instance.merge_error
            command_kwargs["MergeWarning"] = instance.merge_warning
            command_kwargs["MergeVerbose"] = instance.merge_verbose
            command_kwargs["MergeDebug"] = instance.merge_debug

        if protocol_version >= ProtocolVersion.Pwsh5.value:
            command_kwargs["MergeInformation"] = instance.merge_information

        return dict_to_psobject(**command_kwargs)



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import datetime
import decimal
import enum
import queue
import re
import uuid
import xml.etree.ElementTree as ElementTree

import pytest

import psrpcore.types._serializer as serializer
from psrpcore.types import (
    PSBool,
    PSByte,
    PSByteArray,
    PSChar,
    PSCustomObject,
    PSDateTime,
    PSDecimal,
    PSDouble,
    PSDuration,
    PSGuid,
    PSInt,
    PSInt16,
    PSInt64,
    PSQueue,
    PSSByte,
    PSScriptBlock,
    PSSingle,
    PSString,
    PSUInt,
    PSUInt16,
    PSUInt64,
    PSUri,
    PSVersion,
    PSXml,
)

from ..conftest import (
    COMPLEX_ENCODED_STRING,
    COMPLEX_STRING,
    FakeCryptoProvider,
    assert_xml_diff,
)

# A lot of the serializer tests are done in the tests for each object, these are just for extra edge cases we want to
# validate


@pytest.mark.parametrize(
    "input_value, expected",
    [
        (PSBool(True), "<B>true</B>"),
        (PSBool(False), "<B>false</B>"),
        (True, "<B>true</B>"),
        (False, "<B>false</B>"),
        (PSByte(1), "<By>1</By>"),
        (PSByteArray(b"\x00\x01\x02\x03"), "<BA>AAECAw==</BA>"),
        (b"\x00\x01\x02\x03", "<BA>AAECAw==</BA>"),
        (PSChar("a"), "<C>97</C>"),
        (PSDateTime(1970, 1, 1), "<DT>1970-01-01T00:00:00Z</DT>"),
        (datetime.datetime(1970, 1, 1), "<DT>1970-01-01T00:00:00Z</DT>"),
        (PSDecimal(1), "<D>1</D>"),
        (decimal.Decimal(1), "<D>1</D>"),
        (PSDouble(1.0), "<Db>1.0</Db>"),
        (PSDuration(1), "<TS>P1D</TS>"),
        (datetime.timedelta(1), "<TS>P1D</TS>"),
        (PSGuid(int=0), "<G>00000000-0000-0000-0000-000000000000</G>"),
        (uuid.UUID(int=0), "<G>00000000-0000-0000-0000-000000000000</G>"),
        (PSInt(1), "<I32>1</I32>"),
        (1, "<I32>1</I32>"),
        (PSInt16(1), "<I16>1</I16>"),
        (PSInt64(1), "<I64>1</I64>"),
        (PSSingle(1.0), "<Sg>1.0</Sg>"),
        (float(1.0), "<Sg>1.0</Sg>"),
        (PSSByte(1), "<SB>1</SB>"),
        (PSScriptBlock(COMPLEX_STRING), f"<SBK>{COMPLEX_ENCODED_STRING}</SBK>"),
        (PSString(COMPLEX_STRING), f"<S>{COMPLEX_ENCODED_STRING}</S>"),
        (COMPLEX_STRING, f"<S>{COMPLEX_ENCODED_STRING}</S>"),
        (PSUInt(1), "<U32>1</U32>"),
        (PSUInt16(1), "<U16>1</U16>"),
        (PSUInt64(1), "<U64>1</U64>"),
        (PSUri(COMPLEX_STRING), f"<URI>{COMPLEX_ENCODED_STRING}</URI>"),
        (PSVersion("1.2.3.4"), "<Version>1.2.3.4</Version>"),
        (PSXml(COMPLEX_STRING), f"<XD>{COMPLEX_ENCODED_STRING}</XD>"),
    ],
)
def test_serialize_primitive_object(input_value, expected):
    element = serializer.serialize(input_value, FakeCryptoProvider())
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == expected


@pytest.mark.parametrize(
    "input_value, expected",
    [
        ("<B>true</B>", PSBool(True)),
        ("<B>false</B>", PSBool(False)),
        ("<By>1</By>", PSByte(1)),
        ("<BA>AAECAw==</BA>", PSByteArray(b"\x00\x01\x02\x03")),
        ("<C>97</C>", PSChar("a")),
        (
            "<DT>2008-04-11T10:42:32.2731993-07:00</DT>",
            PSDateTime(
                2008,
                4,
                11,
                10,
                42,
                32,
                273199,
                tzinfo=datetime.timezone(-datetime.timedelta(seconds=25200)),
                nanosecond=300,
            ),
        ),
        ("<D>1</D>", PSDecimal(1)),
        ("<Db>1.0</Db>", PSDouble(1.0)),
        ("<TS>PT9.0269026S</TS> ", PSDuration(seconds=9, microseconds=26902, nanoseconds=600)),
        ("<G>00000000-0000-0000-0000-000000000000</G>", PSGuid(int=0)),
        ("<I32>1</I32>", PSInt(1)),
        ("<I16>1</I16>", PSInt16(1)),
        ("<I64>1</I64>", PSInt64(1)),
        ("<Sg>1.0</Sg>", PSSingle(1.0)),
        ("<SB>1</SB>", PSSByte(1)),
        (f"<SBK>{COMPLEX_ENCODED_STRING}</SBK>", PSScriptBlock(COMPLEX_STRING)),
        (f"<S>{COMPLEX_ENCODED_STRING}</S>", PSString(COMPLEX_STRING)),
        ("<U32>1</U32>", PSUInt(1)),
        ("<U16>1</U16>", PSUInt16(1)),
        ("<U64>1</U64>", PSUInt64(1)),
        (f"<URI>{COMPLEX_ENCODED_STRING}</URI>", PSUri(COMPLEX_STRING)),
        ("<Version>1.2.3.4</Version>", PSVersion("1.2.3.4")),
        (f"<XD>{COMPLEX_ENCODED_STRING}</XD>", PSXml(COMPLEX_STRING)),
    ],
)
def test_deserialize_primitive_object(input_value, expected):
    element = ElementTree.fromstring(input_value)
    actual = serializer.deserialize(element, FakeCryptoProvider())
    assert isinstance(actual, type(expected))
    assert actual == expected


def test_deserialize_invalid_duration():
    expected = re.escape("Duration input 'invalid' is not valid, cannot deserialize")
    with pytest.raises(ValueError, match=expected):
        serializer.deserialize(ElementTree.fromstring("<TS>invalid</TS>"), FakeCryptoProvider())


def test_serialize_python_class():
    class MyClass:
        def __init__(self):
            self.attribute = "abc"
            self.__private = "wont appear"

        @property
        def property(self):
            return "def"

        def __str__(self):
            return "MyClass"

        def function(self):
            return "wont appear"

    element = serializer.serialize(MyClass(), FakeCryptoProvider())
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Management.Automation.PSCustomObject</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<MS>"
        '<S N="attribute">abc</S>'
        '<S N="property">def</S>'
        "</MS>"
        "</Obj>"
    )


def test_deserialize_unknown_tag():
    expected = re.escape("Unknown element found: bad")
    with pytest.raises(ValueError, match=expected):
        serializer.deserialize(ElementTree.fromstring("<bad>test</bad>"), FakeCryptoProvider())


def test_deserialize_special_queue():
    clixml = (
        '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Collections.Generic.Queue`1[[System.Object]]</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<QUE>"
        "<I32>1</I32>"
        "<I32>2</I32>"
        "</QUE>"
        "</Obj>"
    )

    actual = serializer.deserialize(ElementTree.fromstring(clixml), FakeCryptoProvider())
    assert actual.PSTypeNames == [
        "Deserialized.System.Collections.Generic.Queue`1[[System.Object]]",
        "Deserialized.System.Object",
    ]
    assert isinstance(actual, PSQueue)
    assert actual.get() == 1
    assert actual.get() == 2
    with pytest.raises(queue.Empty):
        actual.get(block=False)


def test_serialize_native_enum():
    class MyEnum(enum.IntEnum):
        none = 0
        test1 = 1

    element = serializer.serialize(MyEnum.none, FakeCryptoProvider())
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == (
        '<Obj RefId="0">'
        "<I32>0</I32>"
        '<TN RefId="0">'
        f"<T>{MyEnum.__module__}.{MyEnum.__name__}</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>none</ToString>"
        "</Obj>"
    )

    element = serializer.serialize(MyEnum.test1, FakeCryptoProvider())
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == (
        '<Obj RefId="0">'
        "<I32>1</I32>"
        '<TN RefId="0">'
        f"<T>{MyEnum.__module__}.{MyEnum.__name__}</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>test1</ToString>"
        "</Obj>"
    )


def test_serialize_native_flags():
    class MyEnum(enum.IntFlag):
        none = 0
        test1 = 1
        test2 = 2

    element = serializer.serialize(MyEnum.none, FakeCryptoProvider())
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == (
        '<Obj RefId="0">'
        "<I32>0</I32>"
        '<TN RefId="0">'
        f"<T>{MyEnum.__module__}.{MyEnum.__name__}</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>none</ToString>"
        "</Obj>"
    )

    element = serializer.serialize(MyEnum.test1 | MyEnum.test2, FakeCryptoProvider())
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == (
        '<Obj RefId="0">'
        "<I32>3</I32>"
        '<TN RefId="0">'
        f"<T>{MyEnum.__module__}.{MyEnum.__name__}</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>test1, test2</ToString>"
        "</Obj>"
    )


def test_fail_deserialize_dict_no_key():
    clixml = (
        '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Collections.Hashtable</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<DCT>"
        '<En><S N="NotKey">1</S><S N="Value">test</S></En>'
        "</DCT>"
        "</Obj>"
    )

    with pytest.raises(ValueError, match="Failed to find dict Key attribute"):
        serializer.deserialize(ElementTree.fromstring(clixml), FakeCryptoProvider())


def test_fail_deserialize_dict_no_value():
    clixml = (
        '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Collections.Hashtable</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<DCT>"
        '<En><S N="Key">1</S><S N="NotValue">test</S></En>'
        "</DCT>"
        "</Obj>"
    )

    with pytest.raises(ValueError, match="Failed to find dict Value attribute"):
        serializer.deserialize(ElementTree.fromstring(clixml), FakeCryptoProvider())


def test_serialize_circular_reference():
    obj = PSCustomObject(MyProp=1, List=[1], Dict={"test": 1}, CircularRef=None)
    obj.List.append(obj)
    obj.Dict["obj"] = obj
    obj.CircularRef = obj

    element = serializer.serialize(obj, FakeCryptoProvider())
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    expected = (
        '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Management.Automation.PSCustomObject</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<MS>"
        '<I32 N="MyProp">1</I32>'
        '<Obj RefId="1" N="List">'
        '<TN RefId="1">'
        "<T>System.Collections.ArrayList</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<LST>"
        "<I32>1</I32>"
        '<Ref RefId="0" />'
        "</LST>"
        "</Obj>"
        '<Obj RefId="2" N="Dict">'
        '<TN RefId="2">'
        "<T>System.Collections.Hashtable</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<DCT>"
        '<En><S N="Key">test</S><I32 N="Value">1</I32></En>'
        '<En><S N="Key">obj</S><Ref RefId="0" N="Value" /></En>'
        "</DCT>"
        "</Obj>"
        '<Ref RefId="0" N="CircularRef" />'
        "</MS>"
        "</Obj>"
    )
    assert_xml_diff(actual, expected)

    obj = serializer.deserialize(element, FakeCryptoProvider())
    assert isinstance(obj, PSCustomObject)
    assert obj.MyProp == 1
    assert len(obj.List) == 2
    assert obj.List[0] == 1
    assert obj.List[1] == obj
    assert obj.Dict["test"] == 1
    assert obj.Dict["obj"] == obj
    assert obj.CircularRef == obj



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import base64
import functools
import typing
import uuid

from psrpcore._base import Pipeline, RunspacePool
from psrpcore._command import Command
from psrpcore._crypto import create_keypair, decrypt_session_key, rsa
from psrpcore._events import (
    ApplicationPrivateDataEvent,
    DebugRecordEvent,
    EncryptedSessionKeyEvent,
    ErrorRecordEvent,
    InformationRecordEvent,
    PipelineHostCallEvent,
    PipelineHostResponseEvent,
    PipelineOutputEvent,
    PipelineStateEvent,
    ProgressRecordEvent,
    PublicKeyRequestEvent,
    RunspaceAvailabilityEvent,
    RunspacePoolHostCallEvent,
    RunspacePoolHostResponseEvent,
    RunspacePoolInitDataEvent,
    RunspacePoolStateEvent,
    SetRunspaceAvailabilityEvent,
    UserEventEvent,
    VerboseRecordEvent,
    WarningRecordEvent,
)
from psrpcore._exceptions import (
    InvalidPipelineState,
    InvalidProtocolVersion,
    InvalidRunspacePoolState,
)
from psrpcore._payload import ProtocolVersion, StreamType
from psrpcore._pipeline import GetMetadata, PowerShell
from psrpcore.types import (
    ApartmentState,
    CommandTypes,
    ConnectRunspacePool,
    EndOfPipelineInput,
    ErrorRecord,
    GetAvailableRunspaces,
    HostInfo,
    InitRunspacePool,
    PipelineHostResponse,
    PSInvocationState,
    PSRPMessageType,
    PSThreadOptions,
    PublicKey,
    RemoteStreamOptions,
    ResetRunspaceState,
    RunspacePoolHostResponse,
    RunspacePoolState,
    SetMaxRunspaces,
    SetMinRunspaces,
)


class ClientRunspacePool(RunspacePool["_ClientPipeline"]):
    """Client Runspace Pool.

    Represents a Runspace Pool on a remote host which can contain one or more
    running pipelines. This is a non blocking connection object that handles
    the incoming and outgoing PSRP packets without worrying about the IO. This
    model is inspired by `Sans-IO model`_ where this object deals with only
    the PSRP protocol and needs to be combined with an IO transport separately.

    This is meant to be a close representation of the
    `System.Management.Automation.Runspaces.RunspacePool`_ .NET class.

    Args:
        application_arguments: Arguments that are sent to the server and
            accessible through `$PSSenderInfo.ApplicationArguments` of a
            pipeline that runs in this Runspace Pool.
        apartment_state: The apartment state of the thread used to execute
            commands within this Runspace Pool.
        host: The HostInfo that describes the client hosting application.
        thread_options: Determines whether a new thread is created for each
            invocation.
        min_runspaces: The minimum number of Runspaces a pool can hold.
        max_runspaces: The maximum number of Runspaces a pool can hold.
        runspace_pool_id: Manually set the Runspace Pool ID, used when
            reconnecting to an existing Runspace Pool.

    .. _System.Management.Automation.Runspaces.RunspacePool:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.runspaces.runspacepool

    .. _Sans-IO model:
        https://sans-io.readthedocs.io/
    """

    def __init__(
        self,
        application_arguments: typing.Optional[typing.Dict[str, typing.Any]] = None,
        apartment_state: ApartmentState = ApartmentState.Unknown,
        host: typing.Optional[HostInfo] = None,
        thread_options: PSThreadOptions = PSThreadOptions.Default,
        min_runspaces: int = 1,
        max_runspaces: int = 1,
        runspace_pool_id: typing.Optional[uuid.UUID] = None,
    ) -> None:
        super().__init__(
            runspace_pool_id or uuid.uuid4(),
            application_arguments=application_arguments or {},
            application_private_data={},
        )
        self.apartment_state = apartment_state
        self.host = host
        self.thread_options = thread_options
        self._min_runspaces = min_runspaces
        self._max_runspaces = max_runspaces
        self._exchange_key: typing.Optional[rsa.RSAPrivateKey] = None

    def open(self) -> None:
        """Opens the Runspace Pool.

        This opens the Runspace Pool on the peer.
        """
        if self.state == RunspacePoolState.Opened:
            return
        if self.state != RunspacePoolState.BeforeOpen:
            raise InvalidRunspacePoolState("open a Runspace Pool", self.state, [RunspacePoolState.BeforeOpen])

        host = self.host or HostInfo()
        self._change_state(RunspacePoolState.Opening)

        self.prepare_message(self.our_capability)
        self.prepare_message(
            InitRunspacePool(
                MinRunspaces=self._min_runspaces,
                MaxRunspaces=self._max_runspaces,
                PSThreadOptions=self.thread_options,
                ApartmentState=self.apartment_state,
                HostInfo=host,
                ApplicationArguments=self.application_arguments,
            )
        )

    def connect(self) -> None:
        """Connects to the Runspace Pool.

        This connects to a disconnected Runspace Pool on the peer.
        """
        if self.state == RunspacePoolState.Opened:
            return
        if self.state != RunspacePoolState.BeforeOpen and self.state != RunspacePoolState.Disconnected:
            raise InvalidRunspacePoolState("connect to Runspace Pool", self.state, [RunspacePoolState.BeforeOpen])

        self._change_state(RunspacePoolState.Connecting)
        self.prepare_message(self.our_capability)
        self.prepare_message(ConnectRunspacePool())

    def get_available_runspaces(self) -> int:
        """Get the number of Runspaces available.

        This builds a request to get the number of available Runspaces in the
        pool. The :class:`psrpcore.GetRunspaceAvailabilityEvent` event is
        returned once the response is received from the server.

        Returns:
            int: The command id for this request.
        """
        if self.state != RunspacePoolState.Opened:
            raise InvalidRunspacePoolState("get available Runspaces", self.state, [RunspacePoolState.Opened])

        ci = self._ci_counter
        self._ci_handlers[ci] = None
        self.prepare_message(GetAvailableRunspaces(ci=ci))

        return ci

    def exchange_key(self) -> None:
        """Exchange session specific key.

        Request the session key from the peer.
        """
        if self._key_requested:
            return

        if self.state != RunspacePoolState.Opened:
            raise InvalidRunspacePoolState("start session key exchange", self.state, [RunspacePoolState.Opened])

        self._exchange_key, public_key = create_keypair()
        b64_public_key = base64.b64encode(public_key).decode()

        self.prepare_message(PublicKey(PublicKey=b64_public_key))
        self._key_requested = True

    def host_response(
        self,
        ci: int,
        return_value: typing.Optional[typing.Any] = None,
        error_record: typing.Optional[ErrorRecord] = None,
    ) -> None:
        """Respond to a host call.

        Respond to a host call event with either a return value or an error
        record. It is recommended to use :class:`psrpcore.ClientHostResponder`
        to respond to host calls as it will format the return values from a
        .NET type

        Args:
            ci: The call ID associated with the host call to response to.
            return_value: The return value for the host call.
            error_record: The error record raised by the host when running the
                host call.
        """
        if self.state != RunspacePoolState.Opened:
            raise InvalidRunspacePoolState("respond to host call", self.state, [RunspacePoolState.Opened])

        call_event = typing.cast(
            typing.Union[RunspacePoolHostResponseEvent, PipelineHostResponseEvent], self._ci_events[ci]
        )
        method_identifier = call_event.method_identifier
        pipeline_id = call_event.pipeline_id

        host_call_obj = PipelineHostResponse if pipeline_id else RunspacePoolHostResponse

        host_call = host_call_obj(ci=ci, mi=method_identifier)
        if return_value is not None:
            host_call.mr = return_value

        if error_record is not None:
            host_call.me = error_record

        self.prepare_message(host_call, pipeline_id=pipeline_id, stream_type=StreamType.prompt_response)

        # Any of the above may fail causing the ci record to be dropped.
        # Instead only remove it once a response has been serialized and is
        # ready to send.
        del self._ci_events[ci]

    def reset_runspace_state(self) -> int:
        """Reset the Runspace Pool state.

        Resets the variable table for the Runspace Pool back to the default
        state. This requires the server to be running with PowerShell v5 or
        newer. Older versions will fail with :class:`InvalidProtocolVersion`.

        Returns:
            int: The command id for this request.
        """
        their_version = getattr(self.their_capability, "protocolversion", ProtocolVersion.Win7RC.value)
        required_version = ProtocolVersion.Pwsh5.value
        if their_version < required_version:
            raise InvalidProtocolVersion("reset Runspace Pool state", their_version, required_version)

        if self.state != RunspacePoolState.Opened:
            raise InvalidRunspacePoolState("reset Runspace Pool state", self.state, [RunspacePoolState.Opened])

        ci = self._ci_counter
        self._ci_handlers[ci] = None
        self.prepare_message(ResetRunspaceState(ci=ci))

        return ci

    def set_max_runspaces(
        self,
        value: int,
    ) -> typing.Optional[int]:
        """Set the maximum number of runspaces.

        Build a request to set the maximum number of Runspaces the pool
        maintains. The `max_runspaces` property is updated once the
        :class:`SetMaxRunspacesEvent` is received.

        Args:
            value: The maximum number of runspaces in a pool to change to.

        Returns:
            Optional[int]: The command id for this request or ``None`` if no
                request was created.
        """
        if self.state == RunspacePoolState.BeforeOpen or self._max_runspaces == value:
            self._max_runspaces = value
            return None

        ci = self._ci_counter
        self._ci_handlers[ci] = functools.partial(self._set_runspaces_handler, "max", value)
        self.prepare_message(SetMaxRunspaces(MaxRunspaces=value, ci=ci))

        return ci

    def set_min_runspaces(
        self,
        value: int,
    ) -> typing.Optional[int]:
        """Set the minimum number of runspaces.

        Build a request to set the minimum number of Runspaces the pool
        maintains. The `min_runspaces` property is updated once the
        :class:`SetMinRunspacesEvent` is fired.

        Args:
            value: The minimum number of runspaces in a pool to change to.

        Returns:
            Optional[int]: The command id for this request or ``None`` if no
                request was created.
        """
        if self.state == RunspacePoolState.BeforeOpen or self._min_runspaces == value:
            self._min_runspaces = value
            return None

        ci = self._ci_counter
        self._ci_handlers[ci] = functools.partial(self._set_runspaces_handler, "min", value)
        self.prepare_message(SetMinRunspaces(MinRunspaces=value, ci=ci))

        return ci

    def _process_ApplicationPrivateData(
        self,
        event: ApplicationPrivateDataEvent,
    ) -> None:
        self.application_private_data = event.data
        if self.state == RunspacePoolState.Connecting:
            self._change_state(RunspacePoolState.Opened)

    def _process_DebugRecord(
        self,
        event: DebugRecordEvent,
    ) -> None:
        pass

    def _process_EncryptedSessionKey(
        self,
        event: EncryptedSessionKeyEvent,
    ) -> None:
        session_key = decrypt_session_key(
            self._exchange_key,  # type: ignore[arg-type] # Before we get this message the exchange_key is set
            event.key,
        )
        self._cipher.register_key(session_key)

    def _process_ErrorRecord(
        self,
        event: ErrorRecordEvent,
    ) -> None:
        pass

    def _process_InformationRecord(
        self,
        event: InformationRecordEvent,
    ) -> None:
        pass

    def _process_PipelineHostCall(
        self,
        event: PipelineHostCallEvent,
    ) -> None:
        if event.ci != -100:  # Used by pwsh to indicate this is a void method
            self._ci_events[event.ci] = event

    def _process_PipelineOutput(
        self,
        event: PipelineOutputEvent,
    ) -> None:
        pass

    def _process_PipelineState(
        self,
        event: PipelineStateEvent,
    ) -> None:
        pipeline_id = event.pipeline_id
        pipeline = self.pipeline_table[pipeline_id]
        pipeline.state = event.state

    def _process_ProgressRecord(
        self,
        event: ProgressRecordEvent,
    ) -> None:
        pass

    def _process_PublicKeyRequest(
        self,
        event: PublicKeyRequestEvent,
    ) -> None:
        self.exchange_key()

    def _process_RunspaceAvailability(
        self,
        event: RunspaceAvailabilityEvent,
    ) -> None:
        handler = self._ci_handlers.pop(event.ci)
        if handler is not None:
            handler(event)

    def _process_RunspacePoolHostCall(
        self,
        event: RunspacePoolHostCallEvent,
    ) -> None:
        if event.ci != -100:  # Used by pwsh to indicate this is a void method
            self._ci_events[event.ci] = event

    def _process_RunspacePoolInitData(
        self,
        event: RunspacePoolInitDataEvent,
    ) -> None:
        self._min_runspaces = event.min_runspaces
        self._max_runspaces = event.max_runspaces

    def _process_RunspacePoolState(
        self,
        event: RunspacePoolStateEvent,
    ) -> None:
        self._change_state(event.state)

    def _process_UserEvent(
        self,
        event: UserEventEvent,
    ) -> None:
        pass

    def _process_VerboseRecord(
        self,
        event: VerboseRecordEvent,
    ) -> None:
        pass

    def _process_WarningRecord(
        self,
        event: WarningRecordEvent,
    ) -> None:
        pass

    def _set_runspaces_handler(
        self,
        field: str,
        value: int,
        event: SetRunspaceAvailabilityEvent,
    ) -> None:
        """Changes the runspace count based on the result."""
        if event.success:
            setattr(self, f"_{field}_runspaces", value)


class _ClientPipeline(Pipeline["ClientRunspacePool"]):
    """Client Pipeline.

    Represents a client pipeline and the various methods that can be used to
    communicate with the server pipeline. Any data generated by this pipeline
    is retrieved through the Runspace Pool is is a member off.
    """

    def __init__(
        self,
        runspace_pool: "ClientRunspacePool",
        pipeline_id: typing.Optional[uuid.UUID] = None,
    ) -> None:
        super().__init__(runspace_pool, pipeline_id or uuid.uuid4())
        self._message_type: typing.Optional[PSRPMessageType] = None

    def start(self) -> None:
        """Invokes the pipeline.

        Starts the pipeline on the server. This creates the request to create
        and queue the pipeline on the server.
        """
        valid_states = [PSInvocationState.NotStarted, PSInvocationState.Stopped, PSInvocationState.Completed]
        if self.state not in valid_states:
            raise InvalidPipelineState("start a pipeline", self.state, valid_states)

        self.prepare_message(self.metadata, message_type=self._message_type)
        self.state = PSInvocationState.Running

    def send(
        self,
        data: typing.Any,
    ) -> None:
        """Send data to the pipeline.

        Sends input data to a running pipeline.

        Args:
            data: The data to send to the pipeline.
        """
        if self.state != PSInvocationState.Running:
            raise InvalidPipelineState("send pipeline input", self.state, [PSInvocationState.Running])

        self.prepare_message(data, message_type=PSRPMessageType.PipelineInput)

    def send_eof(self) -> None:
        """Send EOF to the input pipe.

        Sends the end of input marker to the pipeline to state no more input is
        expected for the pipeline.
        """
        if self.state != PSInvocationState.Running:
            raise InvalidPipelineState("send pipeline input EOF", self.state, [PSInvocationState.Running])

        self.prepare_message(EndOfPipelineInput())

    def host_response(
        self,
        ci: int,
        return_value: typing.Optional[typing.Any] = None,
        error_record: typing.Optional[ErrorRecord] = None,
    ) -> None:
        """Respond to a host call.

        Respond to a host call event with either a return value or an error
        record.

        Args:
            ci: The call ID associated with the host call to response to.
            return_value: The return value for the host call.
            error_record: The error record raised by the host when running the
                host call.
        """
        if self.state != PSInvocationState.Running:
            raise InvalidPipelineState("response to pipeline host call", self.state, [PSInvocationState.Running])

        self.runspace_pool.host_response(ci, return_value, error_record)


class ClientPowerShell(_ClientPipeline):
    """Client PowerShell Pipeline.

    A PowerShell pipeline to be used from the client. This is the main pipeline
    class used by the client to start a pipeline on the server.

    Args:
        runspace_pool: The Runspace Pool the pipeline is a member off.
        add_to_history: Whether to add the pipeline to the history field of the
            runspace.
        apartment_state: The apartment state of the thread that executes the
            pipeline.
        history: The value to use as a historial reference of the pipeline.
        host: The host information to use when executing the pipeline.
        is_nested: Whether the pipeline is nested in another pipeline or not.
        no_input: Whether there is any data to be input into the pipeline.
        remote_stream_options: Whether to add invocation info the the PowerShell
            streams or not.
        redirect_shell_error_to_out: Redirects the global error output pipe to
            the commands error output pipe.
        pipeline_id: Manually set the Pipeline ID, used when reconnecting to an
            existing Pipeline.
    """

    def __init__(
        self,
        runspace_pool: "ClientRunspacePool",
        add_to_history: bool = False,
        apartment_state: ApartmentState = None,
        history: typing.Optional[str] = None,
        host: typing.Optional[HostInfo] = None,
        is_nested: bool = False,
        no_input: bool = True,
        remote_stream_options: RemoteStreamOptions = RemoteStreamOptions.none,
        redirect_shell_error_to_out: bool = True,
        pipeline_id: typing.Optional[uuid.UUID] = None,
    ) -> None:
        super().__init__(runspace_pool=runspace_pool, pipeline_id=pipeline_id)
        self.metadata: PowerShell = PowerShell(
            add_to_history=add_to_history,
            apartment_state=apartment_state or runspace_pool.apartment_state,
            history=history,
            host=host,
            is_nested=is_nested,
            no_input=no_input,
            remote_stream_options=remote_stream_options,
            redirect_shell_error_to_out=redirect_shell_error_to_out,
        )
        self._message_type = PSRPMessageType.CreatePipeline

    def add_argument(
        self,
        value: typing.Any,
    ) -> "ClientPowerShell":
        """Adds argument to the last command.

        Adds a positional argument to the last command in the pipeline.

        Args:
            value: The argument to add.

        Returns:
            ClientPowerShell: itself
        """
        return self.add_parameter(None, value)

    def add_command(
        self,
        cmdlet: typing.Union[str, Command],
        use_local_scope: typing.Optional[bool] = None,
    ) -> "ClientPowerShell":
        """Adds a command

        Adds a command to the current statement.

        Args:
            cmdlet: The command as a string or :class:`Command` object.
            use_local_scope: Whether the command is run under local scope or
                not.

        Returns:
            ClientPowerShell: itself
        """
        if isinstance(cmdlet, str):
            cmdlet = Command(cmdlet, use_local_scope=use_local_scope)

        elif use_local_scope is not None:
            raise TypeError("Cannot set use_local_scope with Command")

        self.metadata.commands.append(cmdlet)
        return self

    def add_parameter(
        self,
        name: typing.Optional[str],
        value: typing.Any = True,
    ) -> "ClientPowerShell":
        """Adds a parameter to the last command.

        Adds a parameter to the last command in the pipeline. A switch parameter
        can either have ``True``, ``False``, or ``None`` as the value with
        ``None`` being equivalent to ``False``.

        Args:
            name: The name of the parameter to add.
            value: The value to set for the parameter.

        Returns:
            ClientPowerShell: itself
        """
        commands = self.metadata.commands
        if not commands:
            raise ValueError(
                "A command is required to add a parameter/argument. A command must be added to the "
                "PowerShell instance first."
            )

        commands[-1].add_parameter(name, value)
        return self

    def add_parameters(
        self,
        **parameters: typing.Any,
    ) -> "ClientPowerShell":
        """Adds parameters to the last command.

        Like :meth:`ClientPowerShell.add_parameter` this adds multiple
        parameters to the last command in the pipeline. The key word arguments
        corresponds to the parameters to add.

        Args:
            parameters: The parameters to add as kwargs.

        Returns:
            ClientPowerShell: itself
        """
        for name, value in parameters.items():
            self.add_parameter(name, value)

        return self

    def add_script(
        self,
        script: str,
        use_local_scope: typing.Optional[bool] = None,
    ) -> "ClientPowerShell":
        """Adds a script.

        Adds a script to the current statement. A script is essentially a
        scriptblock containing the PowerShell code to execute.

        Args:
            script: The script to add.
            use_local_scope: Whether the command is run under local scope or
                not.

        Returns:
            ClientPowerShell: itself
        """
        return self.add_command(Command(script, True, use_local_scope=use_local_scope))

    def add_statement(self) -> "ClientPowerShell":
        """Adds a statement.

        Adds a statement to the pipeline. A statement can be used to separate
        multiple commands and scripts as separate statements rather than
        pipeline commands.

        Returns:
            ClientPowerShell: itself
        """
        commands = self.metadata.commands
        if commands:
            commands[-1].end_of_statement = True

        return self


class ClientGetCommandMetadata(_ClientPipeline):
    """Client Get Command Metadata Pipeline.

    A Get Command Metadata pipeline to be used from the client.

    Args:
        runspace_pool: The Runspace Pool the pipeline is a member off.
        name: List of command names to get the metadata for. Uses ``*`` as a
            wildcard.
        command_type: The type of commands to filter by.
        namespace: Wildcard patterns describbing the command namespace to filter
            by.
        arguments: Extra arguments passed to the higher-layer above PSRP.
        pipeline_id: Manually set the Pipeline ID, used when reconnecting to an
            existing Pipeline.
    """

    def __init__(
        self,
        runspace_pool: "ClientRunspacePool",
        name: typing.Union[str, typing.List[str]],
        command_type: CommandTypes = CommandTypes.All,
        namespace: typing.Optional[typing.List[str]] = None,
        arguments: typing.Optional[typing.List[typing.Any]] = None,
        pipeline_id: typing.Optional[uuid.UUID] = None,
    ) -> None:
        super().__init__(runspace_pool=runspace_pool, pipeline_id=pipeline_id)
        self.metadata: GetMetadata = GetMetadata(
            name=name,
            command_type=command_type,
            namespace=namespace,
            arguments=arguments,
        )
        self._message_type = PSRPMessageType.GetCommandMetadata



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import struct
import typing

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.padding import PKCS7

from psrpcore._exceptions import MissingCipherError
from psrpcore.types import PSCryptoProvider


def create_keypair() -> typing.Tuple[rsa.RSAPrivateKey, bytes]:
    """Create RSA keypair.

    Create an RSA keypair for use with the PSRemoting session key exchange.

    Returns:
        Tuple[rsa.RSAPrivateKey, bytes]: The RSA Private key generated and the
            public key that can be send to the remote PSSession.
    """
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend(),
    )
    public_numbers = private_key.public_key().public_numbers()
    exponent = struct.pack("<I", public_numbers.e)

    b_modulus = bytearray()
    modulus = public_numbers.n
    while modulus:
        val = modulus & 0xFF
        b_modulus.append(val)
        modulus >>= 8

    # The public key bytes follow a set structure defined in MS-PSRP.
    public_key_bytes = b"\x06\x02\x00\x00\x00\xa4\x00\x00\x52\x53\x41\x31\x00\x08\x00\x00" + exponent + bytes(b_modulus)

    return private_key, public_key_bytes


def encrypt_session_key(
    exchange_key: bytes,
    session_key: bytes,
) -> bytes:
    """Encrypt session key.

    Encrypts the PSRemoting session key generated by the server to be send to
    the client.

    Args:
        exchange_key: The public key generated by the client to encrypt the
            session key with.
        session_key: The session key to encrypt.

    Returns:
        bytes: The encrypted session key to send to the remote PSSession.
    """
    # Exchange key contains header information used by MS Crypto but we don't use them here.
    exponent = struct.unpack("<I", exchange_key[16:20])[0]
    b_modulus = exchange_key[20:]
    shift = 0
    modulus = 0
    for b in b_modulus:
        modulus += b << (8 * shift)
        shift += 1

    public_key = rsa.RSAPublicNumbers(exponent, modulus).public_key(default_backend())

    encrypted_key = public_key.encrypt(
        session_key,
        padding.PKCS1v15(),
    )[::-1]
    encrypted_key_bytes = b"\x01\x02\x00\x00\x10\x66\x00\x00\x00\xa4\x00\x00" + encrypted_key

    return encrypted_key_bytes


def decrypt_session_key(
    exchange_key: rsa.RSAPrivateKey,
    encrypted_session_key: bytes,
) -> bytes:
    """Decrypt session key.

    Decrypts the PSRemoting session key generated by the server.

    Args:
        exchange_key: The RSA private key that can decrypt the encrypted
            session key.
        encrypted_session_key: The encrypted session key received from the
            server.

    Returns:
        bytes: The decrypted session key sent from the server.
    """
    # Strip off Win32 Crypto Blob Header and reverse the bytes.
    encrypted_key = encrypted_session_key[12:][::-1]
    decrypted_key = exchange_key.decrypt(encrypted_key, padding.PKCS1v15())

    return decrypted_key


class PSRemotingCrypto(PSCryptoProvider):
    """PSCryptoProvider used by PSRP for serializing SecureStrings."""

    def __init__(self) -> None:
        self._cipher: typing.Optional[Cipher] = None
        self._padding = PKCS7(algorithms.AES.block_size)

    def decrypt(self, value: bytes) -> bytes:
        if not self._cipher:
            raise MissingCipherError()

        decryptor = self._cipher.decryptor()
        b_dec = decryptor.update(value) + decryptor.finalize()

        unpadder = self._padding.unpadder()
        plaintext = unpadder.update(b_dec) + unpadder.finalize()

        return plaintext

    def encrypt(self, value: bytes) -> bytes:
        if not self._cipher:
            raise MissingCipherError()

        padder = self._padding.padder()
        b_padded = padder.update(value) + padder.finalize()

        encryptor = self._cipher.encryptor()
        b_enc = encryptor.update(b_padded) + encryptor.finalize()

        return b_enc

    def register_key(
        self,
        key: bytes,
    ) -> None:
        algorithm = algorithms.AES(key)
        mode = modes.CBC(b"\x00" * 16)  # PSRP doesn't use an IV
        self._cipher = Cipher(algorithm, mode, default_backend())



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import re

import pytest

from psrpcore._base import Pipeline, RunspacePool


def test_fail_to_init_runspace_base():
    expected = re.escape(
        "Type RunspacePool cannot be instantiated; it can be used only as a base class for client/server runspace "
        "pool types."
    )

    with pytest.raises(TypeError, match=expected):
        RunspacePool()


def test_fail_to_init_pipeline():
    expected = re.escape(
        "Type Pipeline cannot be instantiated; it can be used only as a base class for client/server pipeline types."
    )

    with pytest.raises(TypeError, match=expected):
        Pipeline(None)
